# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics

import numpy as np
import scipy as sp

from qics.cones.base import SymCone


class SecondOrder(SymCone):
    r"""A class representing a second order cone

    .. math::

        \mathcal{Q}_{n+1} = \{ (t, x) \in \mathbb{R} \times \mathbb{R}^{n}
        : t \geq \| x \|_2 \}.

    Parameters
    ----------
    n : :obj:`int`
        Dimension of the vector :math:`x`, i.e., how many terms are in the
        Euclidean norm.
    """

    def __init__(self, n):
        self.n = n

        self.nu = 1  # Barrier parameter

        self.dim = [1, n]
        self.type = ["r", "r"]

        # Update flags
        self.grad_updated = False
        self.congr_aux_updated = False
        self.nt_aux_updated = False

        return

    def get_init_point(self, out):
        point = [np.array([[1.0]]), np.zeros((self.n, 1))]

        self.set_point(point, point)

        out[0][:] = point[0]
        out[1][:] = point[1]

        return out

    def get_feas(self):
        self.feas_updated = True

        self.x, self.z = self.primal, self.dual

        # Check that x0 > ||x1||_2
        if _soc_res(self.x) <= 0:
            self.feas = False
            return self.feas

        # Check that z0 > ||z1||_2
        if self.z is not None:
            if _soc_res(self.z) <= 0:
                self.feas = False
                return self.feas

        self.feas = True
        return self.feas

    def get_dual_feas(self):
        self.z = self.dual
        return _soc_res(self.z) > 0

    def get_val(self):
        return -0.5 * np.log(_soc_res(self.x))

    def update_grad(self):
        assert self.feas_updated
        assert not self.grad_updated

        self.x_res = _soc_res(self.x)
        self.x_res_inv = 1 / self.x_res

        self.grad = [-self.x[0] * self.x_res_inv, self.x[1] * self.x_res_inv]

        self.grad_updated = True

    def hess_prod_ip(self, out, H):
        assert self.grad_updated

        (Ht, Hx) = H

        x_res_inv2 = self.x_res_inv * self.x_res_inv
        coeff = 2 * x_res_inv2 * (self.x[0] * Ht - self.x[1].T @ Hx)

        out[0][:] = coeff * self.x[0] - Ht * self.x_res_inv
        out[1][:] = -coeff * self.x[1] + Hx * self.x_res_inv

        return out

    def hess_congr(self, A):
        assert self.grad_updated
        if not self.congr_aux_updated:
            self.congr_aux(A)

        # First term, i.e., 2/(x'Jx)^2 * AJxx'JA
        lhs = self.x[0] * self.At.T - self.x[1].T @ self.Ax.T
        lhs *= np.sqrt(2.0) * self.x_res_inv
        out = lhs.T @ lhs

        # Second term, i.e., -1/(x'Jx) * AJA
        out -= (self.At @ self.At.T) * self.x_res_inv
        out += (self.Ax @ self.Ax.T) * self.x_res_inv

        return out

    def invhess_prod_ip(self, out, H):
        assert self.grad_updated

        (Ht, Hx) = H

        coeff = 2 * (self.x[0] * Ht + self.x[1].T @ Hx)

        out[0][:] = coeff * self.x[0] - Ht * self.x_res
        out[1][:] = coeff * self.x[1] + Hx * self.x_res

        return out

    def invhess_congr(self, A):
        assert self.grad_updated
        if not self.congr_aux_updated:
            self.congr_aux(A)

        # First term, i.e., 2 * Axx'A
        lhs = self.x[0] * self.At.T + self.x[1].T @ self.Ax.T
        lhs *= np.sqrt(2.0)
        out = lhs.T @ lhs

        # Second term, i.e., -(x'Jx) * AJA
        out -= (self.At @ self.At.T) * self.x_res
        out += (self.Ax @ self.Ax.T) * self.x_res

        return out

    def third_dir_deriv_axpy(self, out, H, a=True):
        assert self.grad_updated

        (Ht, Hx) = H

        x_res_inv2 = self.x_res_inv * self.x_res_inv
        x_res_inv3 = self.x_res_inv * x_res_inv2

        # Gradients of f(t, x) = t^2 - x'x
        DPhit = 2 * self.x[0]
        DPhix = -2 * self.x[1]

        DPhitH = Ht * DPhit
        DPhixH = Hx.T @ DPhix

        # Hessians of f(t, x) = t^2 - x'x
        D2PhittH = 2 * Ht
        D2PhixxH = -2 * Hx

        D2PhitHH = Ht * D2PhittH
        D2PhixHH = Hx.T @ D2PhixxH

        # Third order derivatives of barrier
        coeff1 = DPhitH + DPhixH
        coeff2 = x_res_inv2 * (D2PhitHH + D2PhixHH)
        coeff2 -= 2 * x_res_inv3 * coeff1 * coeff1
        coeff2 *= 0.5

        dder3_t = coeff2 * DPhit
        dder3_t += coeff1 * x_res_inv2 * D2PhittH

        dder3_x = coeff2 * DPhix
        dder3_x += coeff1 * x_res_inv2 * D2PhixxH

        out[0][:] += dder3_t * a
        out[1][:] += dder3_x * a

        return out

    # ==========================================================================
    # Functions specific to symmetric cones for NT scaling
    # ==========================================================================
    # We want the NT scaling point w and scaled variable lambda such that
    #     H(w) s = z  <==> lambda := W^-T s = W z
    # where H(w) = W^T W.
    # See: [Section 4.1]https://www.seas.ucla.edu/~vandenbe/publications/coneprog.pdf

    def nt_aux(self):
        assert not self.nt_aux_updated
        self.x_res = _soc_res(self.x)
        self.z_res = _soc_res(self.z)

        self.rt_x_res = np.sqrt(self.x_res)
        self.rt_z_res = np.sqrt(self.z_res)

        x_bar = [self.x[0] / self.rt_x_res, self.x[1] / self.rt_x_res]
        z_bar = [self.z[0] / self.rt_z_res, self.z[1] / self.rt_z_res]

        xz_bar = x_bar[0] * z_bar[0] + x_bar[1].T @ z_bar[1]
        gamma = np.sqrt((1.0 + xz_bar) / 2.0)

        # Scaling point
        # w_bar = (s_bar + J z_bar) / 2*gamma
        self.w_res = self.rt_x_res / self.rt_z_res
        self.rt_w_res = np.sqrt(self.w_res)
        self.w_bar = [
            (x_bar[0] + z_bar[0]) / (2 * gamma),
            (x_bar[1] - z_bar[1]) / (2 * gamma),
        ]
        # w_bar^1/2 = (w_bar + e) / (2 * (w0_bar + 1))^1/2
        self.rt_w_bar = [
            (1 + self.w_bar[0]) / np.sqrt(2 * (self.w_bar[0] + 1)),
            self.w_bar[1] / np.sqrt(2 * (self.w_bar[0] + 1)),
        ]
        # w = (x'Jx / z'Jz)^1/4 * w_bar
        self.w = [self.w_bar[0] * self.rt_w_res, self.w_bar[1] * self.rt_w_res]

        # Scaled variable
        # lambda0_bar = gamma
        # lambda1_bar = ((gamma + z0_bar)*s1_bar + (gamma + s0_bar)*z1_bar)
        #              -----------------------------------------------------
        #                          (s0_bar + z0_bar + 2*gamma)
        temp = (gamma + z_bar[0]) * x_bar[1] + (gamma + x_bar[0]) * z_bar[1]
        self.lmbda_bar = [gamma, temp / (x_bar[0] + z_bar[0] + 2 * gamma)]
        # lambda = ((x'Jx)(z'Jz))^1/4 * lambda_bar
        self.lmbda = [
            self.lmbda_bar[0] * np.sqrt(self.rt_x_res * self.rt_z_res),
            self.lmbda_bar[1] * np.sqrt(self.rt_x_res * self.rt_z_res),
        ]
        self.lmbda_res = _soc_res(self.lmbda)
        self.rt_lmbda_res = np.sqrt(self.lmbda_res)

        self.nt_aux_updated = True

    def nt_prod_ip(self, out, H):
        if not self.nt_aux_updated:
            self.nt_aux()

        (Ht, Hx) = H

        w_res_inv2 = 1 / (self.w_res * self.w_res)
        coeff = 2 * w_res_inv2 * (self.w[0] * Ht - self.w[1].T @ Hx)

        out[0][:] = coeff * self.w[0] - Ht / self.w_res
        out[1][:] = -coeff * self.w[1] + Hx / self.w_res

        return out

    def nt_congr(self, A):
        if not self.nt_aux_updated:
            self.nt_aux()
        if not self.congr_aux_updated:
            self.congr_aux(A)

        # First term, i.e., 2/(w'Jw)^2 * AJww'JA
        lhs = self.w[0] * self.At.T - self.w[1].T @ self.Ax.T
        lhs *= np.sqrt(2.0) / self.w_res
        out = lhs.T @ lhs

        # Second term, i.e., -(w'Jw) * AJA
        out -= (self.At @ self.At.T) / self.w_res
        out += (self.Ax @ self.Ax.T) / self.w_res

        return out

    def invnt_prod_ip(self, out, H):
        if not self.nt_aux_updated:
            self.nt_aux()

        (Ht, Hx) = H

        coeff = 2 * (self.w[0] * Ht + self.w[1].T @ Hx)

        out[0][:] = coeff * self.w[0] - Ht * self.w_res
        out[1][:] = coeff * self.w[1] + Hx * self.w_res

        return out

    def invnt_congr(self, A):
        if not self.nt_aux_updated:
            self.nt_aux()
        if not self.congr_aux_updated:
            self.congr_aux(A)

        # First term, i.e., 2 * Aww'A
        lhs = self.w[0] * self.At.T + self.w[1].T @ self.Ax.T
        lhs *= np.sqrt(2.0)
        out = lhs.T @ lhs

        # Second term, i.e., -(w'Jw) * AJA
        out -= (self.At @ self.At.T) * self.w_res
        out += (self.Ax @ self.Ax.T) * self.w_res

        return out

    def comb_dir(self, out, ds, dz, sigma_mu):
        # Compute the residual for rs where rs is given as the lhs of
        #     Lambda o (W dz + W^-T ds) = -Lambda o Lambda - (W^-T ds_a) o (W dz_a)
        #                                 + sigma * mu * 1
        # which is rearranged into the form H ds + dz = rs, i.e.,
        #     rs := W^-1 [ Lambda \ (-Lambda o Lambda - (W^-T ds_a) o (W dz_a) + sigma*mu 1) ]
        # See: [Section 5.4]https://www.seas.ucla.edu/~vandenbe/publications/coneprog.pdf
        if not self.nt_aux_updated:
            self.nt_aux()

        # Compute (W^-T ds_a) o (W dz_a)
        rtw_ds = 2 * (self.rt_w_bar[0] * ds[0] - self.rt_w_bar[1].T @ ds[1])
        W_ds = [
            (rtw_ds * self.rt_w_bar[0] - ds[0]) / self.rt_w_res,
            (-rtw_ds * self.rt_w_bar[1] + ds[1]) / self.rt_w_res,
        ]

        rt2_dz = 2 * (self.rt_w_bar[0] * dz[0] + self.rt_w_bar[1].T @ dz[1])
        W_dz = [
            (rt2_dz * self.rt_w_bar[0] - dz[0]) * self.rt_w_res,
            (rt2_dz * self.rt_w_bar[1] + dz[1]) * self.rt_w_res,
        ]

        Wds_Wdz = _soc_prod(W_ds, W_dz)

        # Compute -Lambda o Lambda - [ ... ] + sigma*mu I
        lmbda_lmbda = _soc_prod(self.lmbda, self.lmbda)
        rhs = [
            -lmbda_lmbda[0] - Wds_Wdz[0] + sigma_mu,
            -lmbda_lmbda[1] - Wds_Wdz[1],
        ]

        # Compute Lambda \ [ ... ]
        lmbda_rhs = self.lmbda[1].T @ rhs[1]
        temp = self.lmbda_res * rhs[1] + lmbda_rhs * self.lmbda[1]
        work = [
            (self.lmbda[0] * rhs[0] - lmbda_rhs) / self.lmbda_res,
            (-rhs[0] * self.lmbda[1] + temp / self.lmbda[0]) / self.lmbda_res,
        ]

        # Compute W^-1 [ ... ]
        temp = 2 * (self.rt_w_bar[0] * work[0] - self.rt_w_bar[1].T @ work[1])
        out[0][:] = (temp * self.rt_w_bar[0] - work[0]) / self.rt_w_res
        out[1][:] = (-temp * self.rt_w_bar[1] + work[1]) / self.rt_w_res

    def step_to_boundary(self, ds, dz):
        # Compute the maximum step alpha in [0, 1] we can take such that
        #     s + alpha ds >= 0
        #     z + alpha dz >= 0
        # See: [Section 8.3]https://www.seas.ucla.edu/~vandenbe/publications/coneprog.pdf
        if not self.nt_aux_updated:
            self.nt_aux()

        # Compute (W^-T ds_a) and (W dz_a)
        rtw_ds = 2 * (self.rt_w_bar[0] * ds[0] - self.rt_w_bar[1].T @ ds[1])
        W_ds = [
            (rtw_ds * self.rt_w_bar[0] - ds[0]) / self.rt_w_res,
            (-rtw_ds * self.rt_w_bar[1] + ds[1]) / self.rt_w_res,
        ]

        rtw_dz = 2 * (self.rt_w_bar[0] * dz[0] + self.rt_w_bar[1].T @ dz[1])
        W_dz = [
            (rtw_dz * self.rt_w_bar[0] - dz[0]) * self.rt_w_res,
            (rtw_dz * self.rt_w_bar[1] + dz[1]) * self.rt_w_res,
        ]

        # Compute rho = H(lambda)^1/2 W^-T ds_a
        temp = self.lmbda_bar[0] * W_ds[0] - self.lmbda_bar[1].T @ W_ds[1]
        temp2 = (temp + W_ds[0]) / (1 + self.lmbda_bar[0]) * self.lmbda_bar[1]
        rho = [temp / self.rt_lmbda_res, (W_ds[1] - temp2) / self.rt_lmbda_res]

        # Compute sig = H(lambda)^1/2 W dz_a
        temp = self.lmbda_bar[0] * W_dz[0] - self.lmbda_bar[1].T @ W_dz[1]
        temp2 = (temp + W_dz[0]) / (1 + self.lmbda_bar[0]) * self.lmbda_bar[1]
        sig = [temp / self.rt_lmbda_res, (W_dz[1] - temp2) / self.rt_lmbda_res]

        # Maximum step is given by
        #     alpha := 1 / max(0, ||rho1||_2 - rho0, ||sig1||_2 - sig0)
        # Clamp this step between 0 and 1
        rho_step = (rho[0] - np.sqrt(rho[1].T @ rho[1]))[0, 0]
        sig_step = (sig[0] - np.sqrt(sig[1].T @ sig[1]))[0, 0]

        if rho_step >= 0 and sig_step >= 0:
            return 1.0
        else:
            return 1.0 / max(-rho_step, -sig_step)

    # ==========================================================================
    # Auxilliary functions
    # ==========================================================================
    def congr_aux(self, A):
        assert not self.congr_aux_updated

        if sp.sparse.issparse(A):
            A = A.tocsr()

        self.At = A[:, [0]]
        self.Ax = A[:, 1:]

        self.congr_aux_updated = True


def _soc_res(x):
    return (x[0] * x[0] - x[1].T @ x[1])[0, 0]


def _soc_prod(x, y):
    return [x[0] * y[0] + x[1].T @ y[1], x[0] * y[1] + y[0] * x[1]]



# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics

import numpy as np
import scipy as sp

from qics.quantum import i_kr, p_tr


def density_matrix(n, iscomplex=False):
    r"""Generate a random density matrix, i.e., positive semidefinite
    matrix :math:`X\in\mathbb{H}^n` satisfying :math:`\text{tr}[X] = 1`,
    uniformly on the on the Haar measure.

    Parameters
    ----------
    n : :obj:`int`
        Dimension of random density matrix :math:`X`.
    iscomplex : :obj:`bool`, optional
        Whether the matrix is real (``False``) or complex (``True``).
        The default is ``False``.

    Returns
    -------
    :class:`~numpy.ndarray`
        Random density matrix of dimension ``(n, n)``.
    """
    if iscomplex:
        X = np.random.normal(size=(n, n)) + np.random.normal(size=(n, n)) * 1j
    else:
        X = np.random.normal(size=(n, n))
    rho = X @ X.conj().T
    return rho / np.trace(rho)


def pure_density_matrix(n, iscomplex=False):
    r"""Generate a random pure density matrix i.e., rank 1 positive
    semidefinite matrix :math:`X\in\mathbb{H}^n` satisfying
    :math:`\text{tr}[X] = 1`.

    Parameters
    ----------
    n : :obj:`int`
        Dimension of random pure density matrix :math:`X`.
    iscomplex : :obj:`bool`, optional
        Whether the matrix is real (``False``) or complex (``True``).
        The default is ``False``.

    Returns
    -------
    :class:`~numpy.ndarray`
        Random pure density matrix of dimension ``(n, n)``.

    Notes
    -----
    See [1]_ for additional details.

    .. [1] Khatri, S. (2020) "Random Pure States".
           https://sumeetkhatri.com/wp-content/uploads/2020/05/random_pure_states.pdf
    """
    if iscomplex:
        psi = np.random.normal(size=(n)) + np.random.normal(size=(n)) * 1j
    else:
        psi = np.random.normal(size=(n))
    psi /= np.linalg.norm(psi)
    rho = np.outer(psi, psi.conj())
    rho = (rho + rho.conj().T) * 0.5
    return rho


def unitary(n, iscomplex=False):
    r"""Generate a random unitary matrix, i.e., matrix :math:`U` satisfying
    :math:`U^\dagger U = UU^\dagger = \mathbb{I}`, uniformly distributed on
    Haar measure.

    Parameters
    ----------
    n : :obj:`int`
        Dimension of random unitary matrix :math:`U`.
    iscomplex : :obj:`bool`, optional
        Whether the unitary matrix is real (``False``) or complex
        (``True``). The default is ``False``.

    Returns
    -------
    :class:`~numpy.ndarray`
        Random unitary matrix of dimension ``(n, n)``.

    Notes
    -----
    See [1]_ for additional details.

    .. [1] Meckes, E. (2013) "Random Unitary Matrices and Friends".
           https://case.edu/artsci/math/esmeckes/Meckes_SAMSI_Lecture2.pdf

    """
    if iscomplex:
        X = np.random.normal(size=(n, n)) + np.random.normal(size=(n, n)) * 1j
    else:
        X = np.random.normal(size=(n, n))
    U, _ = np.linalg.qr(X)
    return U


def stinespring_operator(nin, nout=None, nenv=None, iscomplex=False):
    r"""Generate a random Stinespring operator, i.e., isometry :math:`V`
    corresponding to a quantum channel :math:`\mathcal{N}(X)=
    \text{tr}_E[VXV^\dagger]`, uniformly distributed on Hilbert-Schmidt
    measure.

    Parameters
    ----------
    nin : :obj:`int`
        Dimension of the input system.
    nout : :obj:`int`, optional
        Dimension of the output system. The default is ``nin``.
    nenv : :obj:`int`, optional
        Dimension of the environment system. The default is ``nout``.
    iscomplex : :obj:`bool`, optional
        Whether the Stinespring operator is real (``False``) or complex
        (``True``). The default is ``False``.

    Returns
    -------
    :class:`~numpy.ndarray`
        Random Stinespring operator of dimension ``(nout*nenv, nin)``

    Notes
    -----
    See [1]_ for additional details.

    .. [1] Kukulski, Ryszard, et al. "Generating random quantum channels."
           Journal of Mathematical Physics 62.6 (2021).
    """
    nout = nout if (nout is not None) else nin
    nenv = nenv if (nenv is not None) else nout
    U = unitary(nout * nenv, iscomplex=iscomplex)
    return U[:, :nin]


def degradable_channel(nin, nout=None, nenv=None, iscomplex=False):
    r"""Generate random degradable channel :math:`\mathcal{N}`, represented
    as a Stinespring isometry :math:`V` such that

    .. math::

        \mathcal{N}(X) = \text{tr}_2[V X V^\dagger]

    Also returns Stinespring isometry :math:`W` corresponding to the
    complementary channel :math:`\mathcal{N}_\text{c}` such that

    .. math::

        \mathcal{N}_\text{c}(X) = \text{tr}_1[V X V^\dagger]
          = \text{tr}_2[W \mathcal{N}(X) W^\dagger].

    Parameters
    ----------
    nin : :obj:`int`
        Dimension of the input system.
    nout : :obj:`int`, optional
        Dimension of the output system. The default is ``nin``.
    nenv : :obj:`int`, optional
        Dimension of the environment system. The default is ``nout``.
    iscomplex : :obj:`bool`, optional
        Whether the Stinespring operators are real (``False``) or complex
        (``True``). The default is ``False``.

    Returns
    -------
    :class:`~numpy.ndarray`
        Stinespring operator :math:`V` of dimension ``(nout*nenv, nin)``
        corresponding to :math:`\mathcal{N}(X)=\text{tr}_2[V X V^\dagger]`.
    :class:`~numpy.ndarray`
        Stinespring operator :math:`W` of dimension ``(nin*nenv, nout)``
        corresponding to :math:`\mathcal{N}_\text{c}(X)=
        \text{tr}_2[W \mathcal{N}(X) W^\dagger]`.

    Notes
    -----
    See [1]_ for additional details.

    .. [1] Cubitt, Toby S., Mary Beth Ruskai, and Graeme Smith.
           "The structure of degradable quantum channels."
           Journal of Mathematical Physics 49.10 (2008).
    """
    nout = nout if (nout is not None) else nin
    nenv = nenv if (nenv is not None) else nout

    assert nenv <= nin
    dtype = np.complex128 if iscomplex else np.float64

    V = np.zeros((nout * nenv, nin), dtype=dtype)  # N Stinespring isometry
    W = np.zeros((nin * nenv, nout), dtype=dtype)  # Ξ Stinespring isometry

    U = unitary(nin, iscomplex=iscomplex)
    for k in range(nout):
        # Generate random vector
        if iscomplex:
            v = np.random.normal(size=(nenv, 1)) + np.random.normal(size=(nenv, 1)) * 1j
        else:
            v = np.random.normal(size=(nenv, 1))
        v /= np.linalg.norm(v)

        # Make Kraus operator and insert into N Stinespring isometry
        K = v @ U[[k], :]
        V[k * nenv : (k + 1) * nenv, :] = K

        # Make Kraus operator and insert into Ξ Stinespring isometry
        W[k::nin, [k]] = v

    return V, W


def choi_operator(nin, nout=None, M=None, iscomplex=False):
    r"""Generate a random Choi operator :math:`\mathcal{J}_\mathcal{N}`
    representing a quantum channel :math:`\mathcal{N}`, i.e.,

    .. math::

        \mathcal{J}_\mathcal{N} = \sum_{i,j}^n
        \mathcal{N}(| i \rangle\langle j |) \otimes | i \rangle\langle j |,

    uniformly distributed on the Hilbert-Schmidt measure.

    Parameters
    ----------
    nin : :obj:`int`
        Dimension of the input system.
    nout : :obj:`int`, optional
        Dimension of the output system. The default is ``nin``.
    M : :obj:`int`, optional
        Dimension used to determine the rank of the random Choi operator.
        The default is ``nout*nin``.
    iscomplex : :obj:`bool`, optional
        Whether the Choi operator is real (``False``) or complex
        (``True``). The default is ``False``.

    Returns
    -------
    :class:`~numpy.ndarray`
        Random Choi operator corresponding to matrix of size
        ``(nout*nin, nout*nin)``.

    Notes
    -----
    See [1]_ for additional details.

    .. [1] Kukulski, Ryszard, et al. "Generating random quantum channels."
           Journal of Mathematical Physics 62.6 (2021).
    """

    nout = nout if (nout is not None) else nin
    M = M if (M is not None) else nout * nin

    # Sample random Wishart ensemble
    nio = nin * nout
    if iscomplex:
        G = np.random.normal(size=(nio, M)) + np.random.normal(size=(nio, M)) * 1j
        G *= np.sqrt(2.0)
    else:
        G = np.random.normal(size=(nio, M))
    W = G @ G.conj().T

    # Obtain normalization required for trace preserving property
    H = p_tr(W, (nout, nin), 0)
    I_H_irt2 = i_kr(sp.linalg.sqrtm(np.linalg.inv(H)), (nout, nin), 0)

    # Return normalized Choi matrix
    J = I_H_irt2 @ W @ I_H_irt2
    return 0.5 * (J + J.conj().T)



# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics

import numpy as np


def p_tr(mat, dims, sys):
    r"""Performs the partial trace on a multipartite state, e.g., for
    a bipartite state with ``dims=(n0, n1)``, this is the unique linear map
    satisfying

    .. math::

        X \otimes Y \mapsto \text{tr}[X] Y,

    if ``sys=0``, or

    .. math::

        X \otimes Y \mapsto \text{tr}[Y] X,

    if ``sys=1``, for all :math:`X,Y\in\mathbb{H}^n`.

    Parameters
    ----------
    mat : :class:`~numpy.ndarray`
        Array of size ``(n0*n1*...*nk-1, n0*n1*...*nk-1)`` represnting a
        matrix defined on :math:`k` subsystems which we want to take the
        partial trace of.
    dims : :obj:`tuple` of :obj:`int`
        The dimensions ``(n0, n1, ..., nk-1)`` of the :math:`k` subsystems.
    sys : :obj:`int` or :obj:`tuple` of :obj:`int`
        Which of the :math:`k` subsystems to trace out. Can define multiple
        subsystems to trace out.

    Returns
    -------
    :class:`~numpy.ndarray`
        The resulting matrix after taking the partial trace. Has dimension
        ``(n0*n1*...*nk-1 / nx, n0*n1*...*nk-1 / nx)`` where ``nx`` is the
        product of the dimensions of the subsystems that have been traced
        out.

    See also
    --------
    i_kr : The Kronecker product with the identity matrix

    Notes
    -----
    This is the adjoint operator of the Kronecker product with the
    identity matrix.
    """

    if isinstance(sys, int):
        sys = [
            sys,
        ]
    if isinstance(sys, tuple) or isinstance(sys, set):
        sys = list(sys)
    not_sys = list(set(range(len(dims))) - set(sys))

    # Sort subsystems so the ones we want to partial trace are at the front
    reordered_dims = sys + not_sys
    reordered_dims = reordered_dims + [k + len(dims) for k in reordered_dims]

    tr_dim = np.prod([dims[k] for k in sys], dtype=int)
    new_dim = np.prod([dims[k] for k in not_sys], dtype=int)

    temp = np.transpose(mat.reshape(*dims, *dims), reordered_dims)
    temp = temp.reshape(tr_dim, new_dim, tr_dim, new_dim)

    return np.trace(temp, axis1=0, axis2=2)


def p_tr_multi(out, mat, dims, sys):
    r"""Performs the partial trace on a list of bipartite matrices.

    Parameters
    ----------
    out : :class:`~numpy.ndarray`
        Preallocated list of matrices to store the output. Has dimension
        ``(p, n0*n1*...*nk-1 / ni, n0*n1*...*nk-1 / ni)`` where ``i`` is
        the system being traced out.
    mat : :class:`~numpy.ndarray`
        Input ``(p, n0*n1*...*nk-1, n0*n1*...*nk-1)`` list of matrices to
        perform the partial trace on.
    dims : tuple[int]
        The dimensions ``(n0, n1, ..., nk)`` of the ``p`` subsystems.
    sys : int or tuple(int)
        Which systems to trace out.
    """

    if isinstance(sys, int):
        sys = [
            sys,
        ]
    if isinstance(sys, tuple):
        sys = list(sys)
    not_sys = list(set(range(len(dims))) - set(sys))

    # Sort subsystems so the ones we want to partial trace are at the front
    reordered_dims = sys + not_sys
    reordered_dims = (
        [0]
        + [k + 1 for k in reordered_dims]
        + [k + 1 + len(dims) for k in reordered_dims]
    )

    tr_dim = np.prod([dims[k] for k in sys], dtype=int)
    new_dim = np.prod([dims[k] for k in not_sys], dtype=int)

    temp = np.transpose(mat.reshape(-1, *dims, *dims), reordered_dims)
    temp = temp.reshape(-1, tr_dim, new_dim, tr_dim, new_dim)

    np.trace(temp, axis1=1, axis2=3, out=out)

    return out


def i_kr(mat, dims, sys):
    r"""Performs Kronecker product between the identity matrix and a given
    matrix, e.g., if we consider a bipartite setup with ``dims=(n0, n1)``,
    this is the unique linear map satisfying

    .. math::

        X \mapsto \mathbb{I} \otimes X,

    if ``sys=0``, or

    .. math::

        X \mapsto X \otimes \mathbb{I},

    if ``sys=1``, for all :math:`X\in\mathbb{H}^n`.

    Parameters
    ----------
    mat : :class:`~numpy.ndarray`
        Array of size ``(n0*n1*...*nk-1 / nx, n0*n1*...*nk-1 / nx)`` to
        apply the Kronecker product to, where ``nx`` is the product of the
        dimensions of the subsystems specified by ``sys``.
    dims : :obj:`tuple` of :obj:`int`
        The dimensions ``(n0, n1, ..., nk-1)`` of the :math:`k` subsystems
        which the output is defined on.
    sys : :obj:`int` or :obj:`tuple` of :obj:`int`
        Which of the :math:`k` subsystems to apply the Kronecker product
        to. Can define multiple subsystems.

    Returns
    -------
    :class:`~numpy.ndarray`
        The resulting ``(n0*n1*...*nk-1, n0*n1*...*nk-1)`` matrix after
        performing the Kronecker product.

    See also
    --------
    p_tr : The partial trace operator

    Notes
    -----
    This is the adjoint operator of the partial trace.
    """

    if isinstance(sys, int):
        sys = [
            sys,
        ]
    if isinstance(sys, tuple):
        sys = list(sys)
    not_sys = list(set(range(len(dims))) - set(sys))

    # Sort subsystems so the ones we want to partial trace are at the front
    reordered_dims = (
        sys + [k + len(dims) for k in sys] + not_sys + [k + len(dims) for k in not_sys]
    )

    N = np.prod(dims)
    new_dims = [dims[k] for k in not_sys] if len(not_sys) > 0 else [1]

    out = np.zeros((N, N), dtype=mat.dtype)
    out_view = np.transpose(out.reshape(*dims, *dims), reordered_dims)

    r = np.meshgrid(*[range(dims[k]) for k in sys])
    r = list(np.array(r).reshape(len(sys), -1))
    out_view[tuple(r + r)] = mat.reshape(*new_dims, *new_dims)

    return out


def i_kr_multi(out, mat, dims, sys):
    r"""Performs Kronecker product between the indentity matrix and a given
    list of matrices.

    Parameters
    ----------
    out : :class:`~numpy.ndarray`
        Preallocated ``(p, n0*n1*...*nk-1, n0*n1*...*nk-1)`` list of
        matrices to store the output.
    mat : :class:`~numpy.ndarray`
        Input matrix to perform the partial trace on. Has dimension
        ``(p, n0*n1*...*nk-1 / ni, n0*n1*...*nk-1 / ni)`` where ``i`` is
        the system being traced out.
    dim : tuple[int]
        The dimensions ``(n0, n1, ..., nk)`` of the subsystems.
    sys : int or tuple(int)
        Which system to Kroneker product should act on.
    """

    if isinstance(sys, int):
        sys = [
            sys,
        ]
    if isinstance(sys, tuple):
        sys = list(sys)
    not_sys = list(set(range(len(dims))) - set(sys))

    # Sort subsystems so the ones we want to partial trace are at the front
    reordered_dims = [k + 1 for k in sys]
    reordered_dims += [k + 1 + len(dims) for k in sys]
    reordered_dims += [0]
    reordered_dims += [k + 1 for k in not_sys]
    reordered_dims += [k + 1 + len(dims) for k in not_sys]

    new_dims = [dims[k] for k in not_sys]

    out.fill(0.0)
    out_view = np.transpose(out.reshape(-1, *dims, *dims), reordered_dims)

    r = np.meshgrid(*[range(dims[k]) for k in sys])
    r = list(np.array(r).reshape(len(sys), -1))
    out_view[tuple(r + r)] = mat.reshape(-1, *new_dims, *new_dims)

    return out


def partial_transpose(mat, dims, sys):
    r"""Performs the partial transpose on a multipartite matrix, e.g., for
    a bipartite state with ``dims=(n0, n1)``, the unique linear map
    satisfying

    .. math::

        X \otimes Y \mapsto X^\top \otimes Y,

    if ``sys=0``, or

    .. math::

        X \otimes Y \mapsto X \otimes Y^\top,

    if ``sys=1``, for all :math:`X,Y\in\mathbb{C}^{n\times n}`.

    Parameters
    ----------
    mat : :class:`~numpy.ndarray`
        Array of size ``(n0*n1*...*nk-1, n0*n1*...*nk-1)`` represnting a
        matrix defined on :math:`k` subsystems which we want to take the
        partial transpose of.
    dims : :obj:`tuple` of :obj:`int`
        The dimensions ``(n0, n1, ..., nk-1)`` of the :math:`k` subsystems.
    sys : :obj:`int`
        Which of the :math:`k` subsystems to transpose.

    Returns
    -------
    :class:`~numpy.ndarray`
        The resulting ``(n0*n1*...*nk-1, n0*n1*...*nk-1)`` matrix after
        performing the partial transpose.
    """
    N = np.prod(dims)

    temp = mat.reshape(*dims, *dims)
    temp = np.swapaxes(temp, sys, sys + len(dims))

    return temp.reshape(N, N)


def swap(mat, dims, sys1, sys2):
    r"""Swaps two systems of a multipartite quantum state, e.g., for a
    bipartite state with ``dims=(n0, n1)``, it is the unique linear maps
    satisfying

    .. math::

        X \otimes Y \mapsto Y \otimes X,

    for all :math:`X,Y\in\mathbb{H}^{n}`.

    Parameters
    ----------
    mat : :class:`~numpy.ndarray`
        Array of size ``(n0*n1*...*nk-1, n0*n1*...*nk-1)`` represnting a
        matrix defined on :math:`k` subsystems which we want to swap the
        subsystems of.
    dims : :obj:`tuple` of :obj:`int`
        The dimensions ``(n0, n1, ..., nk-1)`` of the :math:`k` subsystems.
    sys1 : :obj:`int`
        First of the :math:`k` subsystems to swap.
    sys2 : :obj:`int`
        Second of the :math:`k` subsystems to swap

    Returns
    -------
    :class:`~numpy.ndarray`
        The resulting ``(n0*n1*...*nk-1, n0*n1*...*nk-1)`` matrix after
        performing the swap operator.
    """
    N = np.prod(dims)

    temp = mat.reshape(*dims, *dims)
    temp = np.swapaxes(temp, sys1, sys2)
    temp = np.swapaxes(temp, len(dims) + sys1, len(dims) + sys2)

    return temp.reshape(N, N)



# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics

import numpy as np
import scipy as sp

from qics._utils.linalg import dense_dot_x
from qics.cones.base import Cone, get_central_ray_entr


class ClassEntr(Cone):
    r"""A class representing a (homogenized) classical entropy cone

    .. math::

        \mathcal{CE}_{n} = \text{cl}\{ (t, u, x) \in \mathbb{R} \times
        \mathbb{R}_{++} \times \mathbb{R}^n_{++} : t \geq -u H(u^{-1}x) \},

    where

    .. math::

        H(x) = -\sum_{i=1}^n x_i \log(x_i),

    is the classical (Shannon) entropy function.

    Parameters
    ----------
    n : :obj:`int`
        Dimension of the vector :math:`x`, i.e., how many terms are in the
        classical entropy function.

    See also
    --------
    ClassRelEntr : Classical relative entropy cone
    QuantEntr : (Homogenized) quantum entropy cone

    Notes
    -----
    The epigraph of the classical entropy can be obtained by enforcing the
    linear constraint :math:`u=1`.

    Additionally, the exponential cone

    .. math::

        \mathcal{E}=\{ (x,y,z)\in\mathbb{R}_+\times\mathbb{R}_+
        \times\mathbb{R} : y \geq x \exp(z/x) \},

    can be modelled by realizing that if :math:`(x,y,z)\in\mathcal{E}`,
    then :math:`(-z, y, x)\in\mathcal{CE}_1`.
    """

    def __init__(self, n):
        self.n = n

        self.nu = 2 + self.n  # Barrier parameter

        self.dim = [1, 1, n]
        self.type = ["r", "r", "r"]

        # Update flags
        self.feas_updated = False
        self.grad_updated = False
        self.hess_aux_updated = False
        self.invhess_aux_updated = False
        self.dder3_aux_updated = False
        self.congr_aux_updated = False

        return

    def get_init_point(self, out):
        (t0, u0, x0) = get_central_ray_entr(self.n)

        point = [
            np.array([[t0]]),
            np.array([[u0]]),
            np.ones((self.n, 1)) * x0,
        ]

        self.set_point(point, point)

        out[0][:] = point[0]
        out[1][:] = point[1]
        out[2][:] = point[2]

        return out

    def get_feas(self):
        if self.feas_updated:
            return self.feas

        self.feas_updated = True

        (self.t, self.u, self.x) = self.primal

        # Check that u and x are strictly positive
        if (self.u <= 0) or any(self.x <= 0):
            self.feas = False
            return self.feas

        # Check that t > -u H(x/u) = Σ_i xi log(xi) - (Σ_i xi) log(u)
        self.sum_x = np.sum(self.x)
        self.log_x = np.log(self.x)
        self.log_u = np.log(self.u[0, 0])

        entr_x = self.x.T @ self.log_x
        entr_xu = self.sum_x * self.log_u
        self.z = (self.t - (entr_x - entr_xu))[0, 0]

        self.feas = self.z > 0
        return self.feas

    def get_val(self):
        return -np.log(self.z) - np.sum(self.log_u) - np.sum(self.log_x)

    def update_grad(self):
        assert self.feas_updated
        assert not self.grad_updated

        # Compute gradients of classical entropy
        # D_u H(u, x) = -Σ_i xi / u
        self.ui = np.reciprocal(self.u)
        self.DPhiu = -self.sum_x * self.ui
        # D_x H(u, x) = log(x) + (1 - log(u))
        self.DPhiX = self.log_x + (1.0 - self.log_u)

        # Compute 1 / x
        self.xi = np.reciprocal(self.x)

        # Compute gradient of barrier function
        self.zi = np.reciprocal(self.z)

        self.grad = [
            -self.zi,
            self.zi * self.DPhiu - self.ui,
            self.zi * self.DPhiX - self.xi,
        ]

        self.grad_updated = True

    def hess_prod_ip(self, out, H):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()

        (Ht, Hu, Hx) = H

        # Hessian product of classical entropy
        D2PhiuH = Hu * self.sum_x * self.ui2 - np.sum(Hx) * self.ui
        D2PhixH = -Hu * self.ui + Hx * self.xi

        # Hessian product of barrier function
        out[0][:] = (Ht - Hu * self.DPhiu - Hx.T @ self.DPhiX) * self.zi2
        out[1][:] = -out[0] * self.DPhiu + self.zi * D2PhiuH + Hu * self.ui2
        out[2][:] = -out[0] * self.DPhiX + self.zi * D2PhixH + Hx * self.xi2

        return out

    def hess_congr(self, A):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()
        if not self.congr_aux_updated:
            self.congr_aux(A)

        p = A.shape[0]
        lhs = np.zeros((p, sum(self.dim)))

        work1, work2 = self.work1, self.work2

        # ======================================================================
        # Hessian products with respect to t
        # ======================================================================
        # D2_t F(t, u, x)[Ht, Hu, Hx]
        #         = (Ht - D_u H(u, x)[Hu] - D_x H(u, x)[Hx]) / z^2
        out_t = self.At - self.Au * self.DPhiu[0, 0]
        out_t -= (self.Ax @ self.DPhiX).ravel()
        out_t *= self.zi2

        lhs[:, 0] = out_t

        # ======================================================================
        # Hessian products with respect to u
        # ======================================================================
        # Hessian products for classical entropy
        # D2_uu Phi(u, x) [Hu] =  sum(x) Hu / u^2
        D2PhiuH = self.Huu * self.Au
        # D2_ux Phi(u, x) [Hx] = -sum(Hx) / u
        D2PhiuH += self.Hux * np.sum(self.Ax, axis=1)

        # Hessian product of barrier function
        # D2_u F(t, u, x)[Ht, Hu, Hx]
        #         = -D2_t F(t, u, x)[Ht, Hu, Hx] * D_u H(u, x)
        #           + (D2_uu H(u, x)[Hu] + D2_ux H(u, x)[Hx]) / z
        #           + Hu / u^2
        out_u = -out_t * self.DPhiu
        out_u += D2PhiuH

        lhs[:, 1] = out_u

        # ======================================================================
        # Hessian products with respect to x
        # ======================================================================
        # Hessian products for classical entropy
        # D2_xx Phi(u, x) [Hx] =  Hx / x
        np.multiply(self.Hxx.T, self.Ax, out=work1)
        # D2_xu Phi(u, x) [Hu] = -Hu / u
        work1 += self.Hux * self.Au.reshape(-1, 1)

        # Hessian product of barrier function
        # D2_x F(t, u, x)[Ht, Hu, Hx]
        #         = -D2_t F(t, u, x)[Ht, Hu, Hx] * D_x H(u, x)
        #           + (D2_xu H(u, x)[Hu] + D2_xx H(u, x)[Hx]) / z
        #           + Hx / x^2
        np.outer(out_t, self.DPhiX, out=work2)
        work1 -= work2

        lhs[:, 2:] = work1

        # Multiply A (H A')
        return dense_dot_x(lhs, A.T)

    def invhess_prod_ip(self, out, H):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()
        if not self.invhess_aux_updated:
            self.update_invhessprod_aux()

        (Ht, Hu, Hx) = H

        Wu = Hu + Ht * self.DPhiu
        Wx = Hx + Ht * self.DPhiX

        # Inverse Hessian product of classical entropy
        out_u = self.rho * (Wu - self.Hxx_inv_Hux.T @ Wx)
        out_x = self.Hxx_inv * Wx - out_u * self.Hxx_inv_Hux

        # Inverse Hessian product of barrier function
        out[0][:] = Ht * self.z2 + out_u * self.DPhiu + out_x.T @ self.DPhiX
        out[1][:] = out_u
        out[2][:] = out_x

        return out

    def invhess_congr(self, A):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()
        if not self.invhess_aux_updated:
            self.update_invhessprod_aux()
        if not self.congr_aux_updated:
            self.congr_aux(A)

        # The inverse Hessian product applied on (Ht, Hu, Hx) for the CE barrier
        # is
        #     (u, x) =  M \ (Wu, Wx)
        #         t  =  z^2 Ht + <DPhi(u, x), (u, x)>
        # where (Wu, Wx) = [(Hu, Hx) + Ht DPhi(u, x)]
        #     M = [ (1 + sum(x)/z) / u^2        -1' / zu       ] = [ a  b']
        #         [        -1 / zu         diag(1/zx + 1/x^2)  ]   [ b  D ]
        #
        # To solve linear systems with M, we simplify it by doing block
        # elimination, in which case we get
        #     u = (Wu - b' D^-1 Wx) / (a - b' D^-1 b)
        #     x = D^-1 (Wx - Wu b)

        p = A.shape[0]
        lhs = np.zeros((p, sum(self.dim)))

        # Compute Wu
        Wu = self.Au + self.At * self.DPhiu[0, 0]

        # Compute Wx
        np.outer(self.At, self.DPhiX, out=self.work1)
        self.work1 += self.Ax

        # ======================================================================
        # Inverse Hessian products with respect to u
        # ======================================================================
        # u = (Wu - b' D^-1 Wx) / (a - b' D^-1 b)
        out_u = self.rho * (Wu - (self.work1 @ self.Hxx_inv_Hux).ravel())
        lhs[:, 1] = out_u

        # ======================================================================
        # Inverse Hessian products with respect to x
        # ======================================================================
        # x = D^-1 (Wx - Wu b)
        self.work1 *= self.Hxx_inv.T
        np.outer(out_u, self.Hxx_inv_Hux, out=self.work2)
        self.work1 -= self.work2
        lhs[:, 2:] = self.work1

        # ======================================================================
        # Inverse Hessian products with respect to t
        # ======================================================================
        # t = z^2 Ht + <DH(u, x), (u, x)>
        out_t = self.z2 * self.At
        out_t += out_u * self.DPhiu[0, 0]
        out_t += (self.work1 @ self.DPhiX).ravel()
        lhs[:, 0] = out_t

        # Multiply A (H A')
        return dense_dot_x(lhs, A.T)

    def third_dir_deriv_axpy(self, out, H, a=True):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()
        if not self.dder3_aux_updated:
            self.update_dder3_aux()

        (Ht, Hu, Hx) = H

        Hu2 = Hu * Hu
        Hx2 = Hx * Hx
        sum_Hx = np.sum(Hx)

        chi = (Ht - self.DPhiu * Hu - self.DPhiX.T @ Hx)[0, 0]
        chi2 = chi * chi

        # Classical entropy Hessians
        D2PhiuH = Hu * self.sum_x * self.ui2 - np.sum(Hx) * self.ui
        D2PhixH = -Hu * self.ui + Hx * self.xi

        D2PhiuHH = Hu * D2PhiuH
        D2PhixHH = Hx.T @ D2PhixH

        # Classical entropy third order derivatives
        D3PhiuHH = -2 * Hu2 * self.sum_x * self.ui3
        D3PhiuHH += 2 * Hu * sum_Hx * self.ui2

        D3PhixHH = -Hx2 * self.xi2
        D3PhixHH += Hu2 * self.ui2

        # Third derivatives of barrier
        dder3_t = -2 * self.zi3 * chi2 - self.zi2 * (D2PhixHH + D2PhiuHH)

        dder3_u = -dder3_t * self.DPhiu
        dder3_u -= 2 * self.zi2 * chi * D2PhiuH
        dder3_u += self.zi * D3PhiuHH
        dder3_u -= 2 * Hu2 * self.ui3

        dder3_x = -dder3_t * self.DPhiX
        dder3_x -= 2 * self.zi2 * chi * D2PhixH
        dder3_x += self.zi * D3PhixHH
        dder3_x -= 2 * Hx2 * self.xi3

        out[0][:] += dder3_t * a
        out[1][:] += dder3_u * a
        out[2][:] += dder3_x * a

        return out

    # ==========================================================================
    # Auxilliary functions
    # ==========================================================================
    def congr_aux(self, A):
        assert not self.congr_aux_updated

        if sp.sparse.issparse(A):
            A = A.toarray()

        self.At = A[:, 0]
        self.Au = A[:, 1]
        self.Ax = A[:, 2:]

        self.work1 = np.empty_like(self.Ax)
        self.work2 = np.empty_like(self.Ax)

        self.congr_aux_updated = True

    def update_hessprod_aux(self):
        assert not self.hess_aux_updated
        assert self.grad_updated

        self.zi2 = self.zi * self.zi
        self.ui2 = self.ui * self.ui
        self.xi2 = self.xi * self.xi

        self.Huu = (self.zi * self.sum_x + 1.0) * self.ui2
        self.Hux = -self.zi * self.ui[0, 0]
        self.Hxx = self.zi * self.xi + self.xi2

        self.hess_aux_updated = True

    def update_invhessprod_aux(self):
        assert not self.invhess_aux_updated
        assert self.grad_updated

        self.Hxx_inv = np.reciprocal(self.Hxx)
        self.Hxx_inv_Hux = self.Hxx_inv * self.Hux
        self.rho = 1.0 / (self.Huu - np.sum(self.Hxx_inv_Hux) * self.Hux)[0, 0]

        self.z2 = self.z * self.z

        self.invhess_aux_updated = True

    def update_dder3_aux(self):
        assert not self.dder3_aux_updated
        assert self.hess_aux_updated

        self.zi3 = self.zi * self.zi2
        self.ui3 = self.ui * self.ui2
        self.xi3 = self.xi * self.xi2

        self.dder3_aux_updated = True



# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi
# Based on test_examples.py from CVXOPT by M. Andersen and L. Vandenberghe.

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics

import os
import unittest


class TestExamples(unittest.TestCase):
    def setUp(self):
        dir_path = os.path.dirname(os.path.realpath(__file__))
        self.expath = os.path.normpath(dir_path + "/../examples")

    def exec_example(self, example):
        fname = os.path.join(self.expath, example)
        gdict = dict()
        with open(fname) as f:
            code = compile(f.read(), fname, "exec")
            exec(code, gdict)
        return gdict

    # Semidefinite programs
    def test_sdp_max_cut(self):
        gdict = self.exec_example("sdp/max_cut.py")
        self.assertEqual(gdict["info"]["sol_status"], "optimal")
        self.assertEqual(gdict["info"]["exit_status"], "solved")
        self.assertAlmostEqual(gdict["info"]["p_obj"], -53.4937692, places=6)

    def test_sdp_product(self):
        gdict = self.exec_example("sdp/product.py")
        self.assertEqual(gdict["info"]["sol_status"], "optimal")
        self.assertEqual(gdict["info"]["exit_status"], "solved")
        self.assertAlmostEqual(gdict["info"]["p_obj"], 32.0626922, places=6)

    def test_sdp_diamond_norm(self):
        gdict = self.exec_example("sdp/diamond_norm.py")
        self.assertEqual(gdict["info"]["sol_status"], "optimal")
        self.assertEqual(gdict["info"]["exit_status"], "solved")
        self.assertAlmostEqual(gdict["info"]["p_obj"], -1.1439575, places=6)

    def test_sdp_state_discrimination(self):
        gdict = self.exec_example("sdp/state_discrimination.py")
        self.assertEqual(gdict["info"]["sol_status"], "optimal")
        self.assertEqual(gdict["info"]["exit_status"], "solved")
        self.assertAlmostEqual(gdict["info"]["p_obj"], -1.0000000, places=6)

    def test_sdp_state_fidelity(self):
        gdict = self.exec_example("sdp/state_fidelity.py")
        self.assertEqual(gdict["info"]["sol_status"], "optimal")
        self.assertEqual(gdict["info"]["exit_status"], "solved")
        self.assertAlmostEqual(gdict["info"]["p_obj"], -0.7536086, places=6)

    def test_sdp_optimal_transport(self):
        gdict = self.exec_example("sdp/optimal_transport.py")
        self.assertEqual(gdict["info"]["sol_status"], "optimal")
        self.assertEqual(gdict["info"]["exit_status"], "solved")
        self.assertAlmostEqual(gdict["info"]["p_obj"], 0.3777573, places=6)

    def test_sdp_dps_heirarchy(self):
        gdict = self.exec_example("sdp/dps_heirarchy.py")
        self.assertEqual(gdict["info"]["sol_status"], "pinfeas")
        self.assertEqual(gdict["info"]["exit_status"], "solved")

    # Quantum relative entropy programs
    def test_qrep_qkd_ebBB84(self):
        gdict = self.exec_example("qrep/qkd_ebBB84.py")
        self.assertEqual(gdict["info"]["sol_status"], "optimal")
        self.assertEqual(gdict["info"]["exit_status"], "solved")
        self.assertAlmostEqual(gdict["info"]["p_obj"], 0.1308120, places=6)

    def test_qrep_nearest_correlation(self):
        gdict = self.exec_example("qrep/nearest_correlation.py")
        self.assertEqual(gdict["info"]["sol_status"], "optimal")
        self.assertEqual(gdict["info"]["exit_status"], "solved")
        self.assertAlmostEqual(gdict["info"]["p_obj"], 1.5837876, places=6)

    def test_qrep_rel_entr_entanglement(self):
        gdict = self.exec_example("qrep/rel_entr_entanglement.py")
        self.assertEqual(gdict["info"]["sol_status"], "optimal")
        self.assertEqual(gdict["info"]["exit_status"], "solved")
        self.assertAlmostEqual(gdict["info"]["p_obj"], 0.0048387, places=6)

    def test_qrep_bregman_projection(self):
        gdict = self.exec_example("qrep/bregman_projection.py")
        self.assertEqual(gdict["info"]["sol_status"], "optimal")
        self.assertEqual(gdict["info"]["exit_status"], "solved")
        self.assertAlmostEqual(gdict["info"]["p_obj"], 42.1867694, places=6)

    def test_qrep_cq_channel_capacity(self):
        gdict = self.exec_example("qrep/cq_channel_capacity.py")
        self.assertEqual(gdict["info"]["sol_status"], "optimal")
        self.assertEqual(gdict["info"]["exit_status"], "solved")
        self.assertAlmostEqual(gdict["info"]["p_obj"], -5.0308689, places=6)

    def test_qrep_ea_channel_capacity(self):
        gdict = self.exec_example("qrep/ea_channel_capacity.py")
        self.assertEqual(gdict["info"]["sol_status"], "optimal")
        self.assertEqual(gdict["info"]["exit_status"], "solved")
        self.assertAlmostEqual(gdict["info"]["p_obj"], -0.6931472, places=6)

    def test_qrep_qq_channel_capacity(self):
        gdict = self.exec_example("qrep/qq_channel_capacity.py")
        self.assertEqual(gdict["info"]["sol_status"], "optimal")
        self.assertEqual(gdict["info"]["exit_status"], "solved")
        self.assertAlmostEqual(gdict["info"]["p_obj"], -0.2754992, places=6)

    def test_qrep_ea_rate_distortion(self):
        gdict = self.exec_example("qrep/ea_rate_distortion.py")
        self.assertEqual(gdict["info"]["sol_status"], "optimal")
        self.assertEqual(gdict["info"]["exit_status"], "solved")
        self.assertAlmostEqual(gdict["info"]["p_obj"], 0.5248915, places=6)

    # Noncommutative perspective programs
    def test_ncp_measured_rel_entr(self):
        gdict = self.exec_example("ncp/measured_rel_entr.py")
        self.assertEqual(gdict["info"]["sol_status"], "optimal")
        self.assertEqual(gdict["info"]["exit_status"], "solved")
        self.assertAlmostEqual(gdict["info"]["p_obj"], 0.8299380, places=6)

    def test_ncp_d_opt_design(self):
        gdict = self.exec_example("ncp/d_opt_design.py")
        self.assertEqual(gdict["info"]["sol_status"], "optimal")
        self.assertEqual(gdict["info"]["exit_status"], "solved")
        self.assertAlmostEqual(gdict["info"]["p_obj"], 39.1674844, places=6)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics

# __init__.py

from qics.quantum.entropy import entropy, purify  # noqa
from qics.quantum.operator import p_tr, p_tr_multi  # noqa
from qics.quantum.operator import i_kr, i_kr_multi  # noqa
from qics.quantum.operator import partial_transpose, swap  # noqa

import qics.quantum.random  # noqa



# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics

import numpy as np
import scipy as sp

from qics.cones.base import SymCone


class NonNegOrthant(SymCone):
    r"""A class representing a nonnegative orthant

    .. math::

        \mathbb{R}^n_+ = \{ x \in \mathbb{R}^n : x \geq 0 \}.

    Parameters
    ----------
    n : :obj:`int`
        Dimension of the cone.
    """

    def __init__(self, n):
        self.n = n

        self.nu = n  # Barrier parameter

        self.dim = [n]
        self.type = ["r"]

        self.congr_aux_updated = False

        return

    def get_init_point(self, out):
        self.set_point([np.ones((self.n, 1))], [np.ones((self.n, 1))])

        out[0][:] = self.x

        return out

    def set_point(self, primal, dual=None, a=True):
        self.x = primal[0] * a
        self.z = dual[0] * a if (dual is not None) else None

    def set_dual(self, dual, a=True):
        self.z = dual[0] * a

    def get_feas(self):
        if np.any(np.less_equal(self.x, 0)):
            return False

        if self.z is None and np.any(np.less_equal(self.z, 0)):
            return False

        return True

    def get_dual_feas(self):
        return np.all(np.greater(self.z, 0))

    def get_val(self):
        return -np.sum(np.log(self.x))

    def grad_ip(self, out):
        out[0][:] = -np.reciprocal(self.x)
        return out

    def hess_prod_ip(self, out, H):
        out[0][:] = H[0] / (self.x**2)
        return out

    def hess_congr(self, A):
        return self.base_congr(A, np.reciprocal(self.x))

    def invhess_prod_ip(self, out, H):
        out[0][:] = H[0] * (self.x**2)
        return out

    def invhess_congr(self, A):
        return self.base_congr(A, self.x)

    def base_congr(self, A, x):
        if sp.sparse.issparse(A):
            if not hasattr(self, "Ax"):
                self.Ax = A.copy()
            self.Ax.data = A.data * np.take(x, A.col)
            return self.Ax @ self.Ax.T
        else:
            Ax = x * A.T
            return Ax.T @ Ax

    def third_dir_deriv_axpy(self, out, H, a=True):
        out[0] -= 2 * a * H[0] * H[0] / (self.x * self.x * self.x)
        return out

    def prox(self):
        return np.linalg.norm(self.x * self.z - 1, np.inf)

    # ========================================================================
    # Functions specific to symmetric cones for NT scaling
    # ========================================================================
    # We want the NT scaling point w and scaled variable lambda such that
    #     H(w) s = z  <==> lambda := W^-T s = W z
    # where H(w) = W^T W. For the nonnegative orthant, we have
    #     w      = s.^1/2 ./ z.^1/2
    #     lambda = z.^1/2 .* s.^1/2
    # Also, we have the linear transformations given by
    #     H(w) ds = ds .* s ./ z
    #     W^-T ds = ds .* z.^1/2 ./ s.^1/2
    #     W    dz = dz .* s.^1/2 ./ z.^1/2
    # See: [Section 4.1]https://www.seas.ucla.edu/~vandenbe/publications/coneprog.pdf

    def nt_prod_ip(self, out, H):
        out[0][:] = H[0] * self.z / self.x
        return out

    def nt_congr(self, A):
        return self.base_congr(A, np.sqrt(self.z / self.x))

    def invnt_prod_ip(self, out, H):
        out[0][:] = H[0] * self.x / self.z
        return out

    def invnt_congr(self, A):
        return self.base_congr(A, np.sqrt(self.x / self.z))

    def comb_dir(self, out, ds, dz, sigma_mu):
        # Compute the residual for rs where rs is given as the lhs of
        #     Lambda o (W dz + W^-T ds) = -Lambda o Lambda - (W^-T ds_a) o (W dz_a)
        #                                 + sigma * mu * 1
        # which is rearranged into the form H ds + dz = rs, i.e.,
        #     rs := W^-1 [ Lambda \ (-Lambda o Lambda - (W^-T ds_a) o (W dz_a) + sigma*mu 1) ]
        # See: [Section 5.4]https://www.seas.ucla.edu/~vandenbe/publications/coneprog.pdf
        out[0][:] = (sigma_mu - ds[0] * dz[0]) / self.x - self.z

    def step_to_boundary(self, ds, dz):
        # Compute the maximum step alpha in [0, 1] we can take such that
        #     s + alpha ds >= 0
        #     z + alpha dz >= 0
        # See: [Section 8.3]https://www.seas.ucla.edu/~vandenbe/publications/coneprog.pdf

        # Compute rho := ds / s and sig := dz / z
        min_rho = np.min(ds / self.x)
        min_sig = np.min(dz / self.z)

        # Maximum step is given by
        #     alpha := 1 / max(0, -min(rho), -min(sig))
        # Clamp this step between 0 and 1
        if min_rho >= 0 and min_sig >= 0:
            return 1.0
        else:
            return 1.0 / max(-min_rho, -min_sig)



# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics

import numpy as np
import scipy as sp

from qics._utils.linalg import dense_dot_x
from qics.cones.base import Cone, get_central_ray_relentr


class ClassRelEntr(Cone):
    r"""A class representing a classical relative entropy cone

    .. math::

        \mathcal{CRE}_{n} = \text{cl}\{ (t, x, y) \in \mathbb{R} \times
        \mathbb{R}^n_{++} \times \mathbb{R}^n_{++} : t \geq H(x \| y) \},

    where

    .. math::

        H(x \| y) = \sum_{i=1}^n x_i \log(x_i / y_i),

    is the classical relative entropy function (Kullback-Leibler
    divergence).

    Parameters
    ----------
    n : :obj:`int`
        Dimension of the vectors :math:`x` and :math:`y`, i.e., how many
        terms are in the classical relative entropy function.

    See also
    --------
    ClassEntr : (Homogenized) classical entropy cone
    QuantRelEntr : Quantum relative entropy cone
    """

    def __init__(self, n):
        self.n = n

        self.nu = 1 + 2 * self.n  # Barrier parameter

        self.dim = [1, n, n]
        self.type = ["r", "r", "r"]

        self.idx_X = slice(1, 1 + n)
        self.idx_Y = slice(1 + n, 1 + 2 * n)

        # Update flags
        self.feas_updated = False
        self.grad_updated = False
        self.congr_aux_updated = False
        self.hess_aux_updated = False
        self.invhess_aux_updated = False
        self.dder3_aux_updated = False

        return

    def get_init_point(self, out):
        (t0, x0, y0) = get_central_ray_relentr(self.n)

        point = [
            np.array([[t0]]),
            np.ones((self.n, 1)) * x0,
            np.ones((self.n, 1)) * y0,
        ]

        self.set_point(point, point)

        out[0][:] = point[0]
        out[1][:] = point[1]
        out[2][:] = point[2]

        return out

    def get_feas(self):
        if self.feas_updated:
            return self.feas

        self.feas_updated = True

        (self.t, self.x, self.y) = self.primal

        # Check that x and y are strictly positive
        if any(self.x <= 0) or any(self.y <= 0):
            self.feas = False
            return self.feas

        # Check that t > H(x||y)
        self.log_x = np.log(self.x)
        self.log_y = np.log(self.y)

        self.z = (self.t - (self.x.T @ (self.log_x - self.log_y)))[0, 0]

        self.feas = self.z > 0
        return self.feas

    def get_val(self):
        return -np.log(self.z) - np.sum(self.log_x) - np.sum(self.log_y)

    def update_grad(self):
        assert self.feas_updated
        assert not self.grad_updated

        # Compute gradients of classical relative entropy
        # D_x H(x||y) = log(x) - log(y) + 1
        self.DPhiX = self.log_x - self.log_y + 1
        # D_y H(x||y) = -x / y
        self.DPhiY = -self.x / self.y

        # Compute 1 / x and 1 / y
        self.xi = np.reciprocal(self.x)
        self.yi = np.reciprocal(self.y)

        # Compute gradient of barrier function
        self.zi = np.reciprocal(self.z)

        self.grad = [
            -self.zi,
            self.zi * self.DPhiX - self.xi,
            self.zi * self.DPhiY - self.yi,
        ]

        self.grad_updated = True

    def hess_prod_ip(self, out, H):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()

        (Ht, Hx, Hy) = H

        # Hessian product of classical relative entropy
        D2PhiXH = Hx * self.xi - Hy * self.yi
        D2PhiYH = -Hx * self.yi + Hy * self.x * self.yi2

        # Hessian product of barrier function
        out[0][:] = (Ht - Hx.T @ self.DPhiX - Hy.T @ self.DPhiY) * self.zi2
        out[1][:] = -out[0] * self.DPhiX + self.zi * D2PhiXH + Hx * self.xi2
        out[2][:] = -out[0] * self.DPhiY + self.zi * D2PhiYH + Hy * self.yi2

        return out

    def hess_congr(self, A):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()
        if not self.congr_aux_updated:
            self.congr_aux(A)

        p = A.shape[0]
        lhs = np.empty((p, sum(self.dim)))

        work0, work1 = self.work0, self.work1

        # ======================================================================
        # Hessian products with respect to t
        # ======================================================================
        # D2_t F(t, x, y)[Ht, Hx, Hy]
        #         = (Ht - D_x H(x||y)[Hx] - D_y H(x||y)[Hy]) / z^2
        outt = self.At - (self.Ax @ self.DPhiX).ravel()
        outt -= (self.Ay @ self.DPhiY).ravel()
        outt *= self.zi2

        lhs[:, 0] = outt

        # ======================================================================
        # Hessian products with respect to x
        # ======================================================================
        # Precompute Hessian products for classical relative entropy
        # D2_xx Phi(x, y) [Hx] =  Hx / x
        np.multiply(self.Ax, self.Hxx.T, out=work0)
        # D2_xy Phi(x, y) [Hx] = -Hx / y
        np.multiply(self.Ay, self.Hxy.T, out=work1)

        # Hessian product of barrier function
        # D2_x F(t, x, y)[Ht, Hx, Hy]
        #         = -D2_t F(t, x, y)[Ht, Hx, Hy] * D_x H(x||y)
        #           + (D2_xx H(x||y)[Hx] + D2_xy H(x||y)[Hy]) / z
        #           + Hx / x^2
        work0 += work1
        np.outer(outt, self.DPhiX, out=work1)
        work0 -= work1

        lhs[:, self.idx_X] = work0

        # ======================================================================
        # Hessian products with respect to y
        # ======================================================================
        # Precompute Hessian products for classical relative entropy
        # D2_yx Phi(x, y) [Hy] = -Hy / y
        np.multiply(self.Ax, self.Hxy.T, out=work0)
        # D2_yy Phi(x, y) [Hy] =  Hy * x / y^2
        np.multiply(self.Ay, self.Hyy.T, out=work1)

        # Hessian product of barrier function
        # D2_y F(t, x, y)[Ht, Hx, Hy]
        #         = -D2_t F(t, x, y)[Ht, Hx, Hy] * D_y H(x||y)
        #           + (D2_yx H(x||y)[Hx] + D2_yy H(x||y)[Hy]) / z
        #           + Hy / y^2
        work0 += work1
        np.outer(outt, self.DPhiY, out=work1)
        work0 -= work1

        lhs[:, self.idx_Y] = work0

        # Multiply A (H A')
        return dense_dot_x(lhs, A.T)

    def invhess_prod_ip(self, out, H):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()
        if not self.invhess_aux_updated:
            self.update_invhessprod_aux()

        (Ht, Hx, Hy) = H

        Wx = Hx + Ht * self.DPhiX
        Wy = Hy + Ht * self.DPhiY

        # Inverse Hessian product of classical relative entropy
        outX = self.Hxx_inv * Wx + self.Hxy_inv * Wy
        outY = self.Hxy_inv * Wx + self.Hyy_inv * Wy

        # Inverse Hessian product of barrier function
        out[0][:] = Ht * self.z2 + outX.T @ self.DPhiX + outY.T @ self.DPhiY
        out[1][:] = outX
        out[2][:] = outY

        return out

    def invhess_congr(self, A):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()
        if not self.invhess_aux_updated:
            self.update_invhessprod_aux()
        if not self.congr_aux_updated:
            self.congr_aux(A)

        # The inverse Hessian product applied on (Ht, Hx, Hy) for the CRE
        # barrier is
        #     (x, y) =  M \ (Wx, Wy)
        #         t  =  z^2 Ht + <DPhi(x, y), (x, y)>
        # where (Wx, Wy) = [(Hx, Hy) + Ht DPhi(x, y)]
        #     M = [  diag(1/zx + 1/x^2)       -diag(1/zy)     ] = [ Hxx Hxy ]
        #         [     -diag(1/zy)      diag(x/zy^2 + 1/y^2) ]   [ Hxy Hyy ]
        # The inverse of a block matrix with diagonal blocks is another block
        # matrix with diaognal blocks
        #     M^-1 = [ (Hxx - Hxy^2 * Hyy^-1)^-1  (Hxy - Hxx Hyy Hxy^-1)^-1 ]
        #            [ (Hxy - Hxx Hyy Hxy^-1)^-1  (Hyy - Hxy^2 * Hxx^-1)^-1 ]
        #          = [ Hxx_inv  Hxy_inv ]
        #            [ Hxy_inv  Hyy_inv ]

        p = A.shape[0]
        lhs = np.empty((p, sum(self.dim)))

        work0, work1 = self.work0, self.work1
        work2, work3, work4 = self.work2, self.work3, self.work4

        # Compute Wx = Hx + Ht D_x H(x||y)
        np.outer(self.At, self.DPhiX, out=work4)
        work4 += self.Ax
        # Compute Wy = Hy + Ht D_y H(x||y)
        np.outer(self.At, self.DPhiY, out=work3)
        work3 += self.Ay

        # ======================================================================
        # Inverse Hessian products with respect to x
        # ======================================================================
        # x = Hxx_inv Wx + Hxy_inv Wy
        np.multiply(work4, self.Hxx_inv.T, out=work0)
        np.multiply(work3, self.Hxy_inv.T, out=work1)
        work0 += work1
        lhs[:, self.idx_X] = work0

        # ======================================================================
        # Inverse Hessian products with respect to y
        # ======================================================================
        # y = Hxy_inv Wx + Hyy_inv Wy
        np.multiply(work4, self.Hxy_inv.T, out=work1)
        np.multiply(work3, self.Hyy_inv.T, out=work2)
        work1 += work2
        lhs[:, self.idx_Y] = work1

        # ======================================================================
        # Inverse Hessian products with respect to t
        # ======================================================================
        # t = z^2 Ht + <DPhi(x, y), (x, y)>
        outt = self.z2 * self.At
        outt += (work0 @ self.DPhiX).ravel()
        outt += (work1 @ self.DPhiY).ravel()
        lhs[:, 0] = outt

        return dense_dot_x(lhs, A.T)

    def third_dir_deriv_axpy(self, out, H, a=True):
        if not self.hess_aux_updated:
            self.update_hessprod_aux()
        if not self.dder3_aux_updated:
            self.update_dder3_aux()

        (Ht, Hx, Hy) = H

        Hx2 = Hx * Hx
        Hy2 = Hy * Hy

        chi = (Ht - self.DPhiX.T @ Hx - self.DPhiY.T @ Hy)[0, 0]
        chi2 = chi * chi

        # Classical relative entropy Hessians
        D2PhiXH = Hx * self.xi - Hy * self.yi
        D2PhiYH = -Hx * self.yi + Hy * self.x * self.yi2

        D2PhiXHH = Hx.T @ D2PhiXH
        D2PhiYHH = Hy.T @ D2PhiYH

        # Classical relative entropy third order derivatives
        D3PhiXHH = -Hx2 * self.xi2 + Hy2 * self.yi2
        D3PhiYHH = 2 * Hy * (Hx - Hy * self.x * self.yi) * self.yi2

        # Third derivatives of barrier
        dder3_t = -2 * self.zi3 * chi2 - self.zi2 * (D2PhiXHH + D2PhiYHH)

        dder3_x = -dder3_t * self.DPhiX
        dder3_x -= 2 * self.zi2 * chi * D2PhiXH
        dder3_x += self.zi * D3PhiXHH
        dder3_x -= 2 * Hx2 * self.xi3

        dder3_y = -dder3_t * self.DPhiY
        dder3_y -= 2 * self.zi2 * chi * D2PhiYH
        dder3_y += self.zi * D3PhiYHH
        dder3_y -= 2 * Hy2 * self.yi3

        out[0][:] += dder3_t * a
        out[1][:] += dder3_x * a
        out[2][:] += dder3_y * a

        return out

    # ==========================================================================
    # Auxilliary functions
    # ==========================================================================
    def congr_aux(self, A):
        assert not self.congr_aux_updated

        if sp.sparse.issparse(A):
            A = A.toarray()

        self.At = A[:, 0]
        self.Ax = np.ascontiguousarray(A[:, self.idx_X])
        self.Ay = np.ascontiguousarray(A[:, self.idx_Y])

        self.work0 = np.empty_like(self.Ax)
        self.work1 = np.empty_like(self.Ax)
        self.work2 = np.empty_like(self.Ax)
        self.work3 = np.empty_like(self.Ax)
        self.work4 = np.empty_like(self.Ax)

        self.congr_aux_updated = True

    def update_hessprod_aux(self):
        assert not self.hess_aux_updated
        assert self.grad_updated

        self.zi2 = self.zi * self.zi
        self.xi2 = self.xi * self.xi
        self.yi2 = self.yi * self.yi

        self.Hxx = self.zi * self.xi + self.xi2
        self.Hxy = -self.zi * self.yi
        self.Hyy = (self.zi * self.x + 1) * self.yi2

        self.hess_aux_updated = True

    def update_invhessprod_aux(self):
        assert not self.invhess_aux_updated
        assert self.grad_updated

        self.z2 = self.z * self.z

        self.Hxx_inv = np.reciprocal(self.Hxx - self.Hxy * self.Hxy / self.Hyy)
        self.Hxy_inv = np.reciprocal(self.Hxy - self.Hxx * self.Hyy / self.Hxy)
        self.Hyy_inv = np.reciprocal(self.Hyy - self.Hxy * self.Hxy / self.Hxx)

    def update_dder3_aux(self):
        assert not self.dder3_aux_updated
        assert self.hess_aux_updated

        self.zi3 = self.zi * self.zi2
        self.xi3 = self.xi * self.xi2
        self.yi3 = self.yi * self.yi2

        self.dder3_aux_updated = True



# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics

import numpy as np
import scipy as sp

from qics._utils.gradient import D1_log, D2_log, scnd_frechet
from qics._utils.linalg import (
    cho_fact,
    cho_solve,
    congr_multi,
    dense_dot_x,
    inp,
    x_dot_dense,
)
from qics.cones.base import Cone
from qics.quantum import i_kr, i_kr_multi, p_tr, p_tr_multi
from qics.vectorize import get_full_to_compact_op, vec_to_mat


class QuantCondEntr(Cone):
    r"""A class representing a quantum conditional entropy cone defined on
    :math:`k` subsystems, where the :math:`i`-th subsystem has dimension
    :math:`n_i`, and the :math:`i`-th subsystem is being traced out, i.e.,

    .. math::

        \mathcal{QCE}_{\{n_i\}, j} = \text{cl}\{ (t, X) \in \mathbb{R}
        \times \mathbb{H}^{\Pi_in_i}_{++} : t \geq -S(X) + S(\text{tr}_i(X)) \},

    where

    .. math::

        S(X) = -\text{tr}[X \log(X)],

    is the quantum (von Neumann) entropy, and :math:`\text{tr}_i` is the
    partial trace on the :math:`i`-th subsystem.

    Parameters
    ----------
    dims : :obj:`tuple` of :obj:`int` or :obj:`list` of :obj:`int`
        List of dimensions :math:`\{ n_i \}_{i=0}^{k-1}` of the :math:`k`
        subsystems.
    sys : :obj:`int` or :obj:`tuple` of :obj:`int` or :obj:`list` of :obj:`int`
        Which systems are being traced out by the partial trace. Can define
        multiple subsystems to trace out.
    iscomplex : :obj:`bool`
        Whether the matrix :math:`X` is defined over :math:`\mathbb{H}^n`
        (``True``), or restricted to :math:`\mathbb{S}^n` (``False``). The
        default is ``False``.

    See also
    --------
    QuantRelEntr : Quantum relative entropy cone

    Notes
    -----
    The quantum conditional entropy can also be modelled by the quantum
    relative entropy by noting the identity

    .. math::

        S(X \| \mathbb{I} \otimes \text{tr}_1(X))
        = -S(X) + S(\text{tr}_1(X)).

    However, the cone oracles for the quantum conditional entropy cone are
    much more efficient than those for the quantum relative entropy cone,
    so it is recommended to use the quantum conditional entropy cone where
    possible.
    """

    def __init__(self, dims, sys, iscomplex=False):
        if isinstance(sys, int):
            sys = [sys]
        if isinstance(sys, tuple) or isinstance(sys, set):
            sys = list(sys)

        assert all(
            [sys_k < len(dims) for sys_k in sys]
        ), "Invalid subsystems specified, exceeds total number of dimensions provided."

        self.dims = dims
        self.sys = sys
        self.iscomplex = iscomplex

        self.N = np.prod(dims)  # Total dim. of multipartite system
        self.m = np.prod([dims[k] for k in sys])  # Dim. of system traced out
        self.n = self.N // self.m  # Dim. of system after partial trace

        self.nu = 1 + self.N  # Barrier parameter

        if iscomplex:
            self.vn = self.n * self.n
            self.dim = [1, 2 * self.N * self.N]
            self.type = ["r", "h"]
            self.dtype = np.complex128
        else:
            self.vn = self.n * (self.n + 1) // 2
            self.dim = [1, self.N * self.N]
            self.type = ["r", "s"]
            self.dtype = np.float64

        # Update flags
        self.feas_updated = False
        self.grad_updated = False
        self.hess_aux_updated = False
        self.invhess_aux_updated = False
        self.invhess_aux_aux_updated = False
        self.congr_aux_updated = False
        self.dder3_aux_updated = False

        return

    def get_iscomplex(self):
        return self.iscomplex

    def get_init_point(self, out):
        a = self.N * np.log(self.m) ** 2 + 1
        b = -(2.0 + (1 + self.N) * self.N * np.log(self.m) ** 2)

        t0 = np.sqrt((-b - np.sqrt(b * b - 4 * a)) / (2 * a))
        x0 = np.sqrt((1 + self.N - t0 * t0) / self.N)

        point = [np.array([[t0]]), np.eye(self.N, dtype=self.dtype) * x0]

        self.set_point(point, point)

        out[0][:] = point[0]
        out[1][:] = point[1]

        return out

    def get_feas(self):
        if self.feas_updated:
            return self.feas

        self.feas_updated = True

        (self.t, self.X) = self.primal
        self.Y = p_tr(self.X, self.dims, self.sys)

        # Check that X and Y are positive definite
        #   Note that Y = pTr(X) should be positive definite if X is,
        #   but check just to be sure that we can safely take logarithms of Y.
        self.Dx, self.Ux = np.linalg.eigh(self.X)
        self.Dy, self.Uy = np.linalg.eigh(self.Y)

        if any(self.Dx <= 0) or any(self.Dy <= 0):
            self.feas = False
            return self.feas

        # Check that t > -S(X) + S(pTr(X))
        self.log_Dx = np.log(self.Dx)
        self.log_Dy = np.log(self.Dy)

        log_X = (self.Ux * self.log_Dx) @ self.Ux.conj().T
        log_X = (log_X + log_X.conj().T) * 0.5
        log_Y = (self.Uy * self.log_Dy) @ self.Uy.conj().T
        log_Y = (log_Y + log_Y.conj().T) * 0.5

        self.log_XY = log_X - i_kr(log_Y, self.dims, self.sys)
        self.z = self.t[0, 0] - inp(self.X, self.log_XY)

        self.feas = self.z > 0
        return self.feas

    def get_val(self):
        return -np.log(self.z) - np.sum(self.log_Dx)

    def update_grad(self):
        assert self.feas_updated
        assert not self.grad_updated

        # Compute gradients of quantum conditional entropy
        # DPhi(X) = log(X) - I ⊗ pTr(X)
        self.DPhi = self.log_XY

        # Compute X^-1
        inv_Dx = np.reciprocal(self.Dx)
        inv_X_rt2 = self.Ux * np.sqrt(inv_Dx)
        self.inv_X = inv_X_rt2 @ inv_X_rt2.conj().T

        # Compute gradient of barrier function
        self.zi = np.reciprocal(self.z)
        self.grad = [-self.zi, self.zi * self.DPhi - self.inv_X]

        self.grad_updated = True

    def hess_prod_ip(self, out, H):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()

        (Ht, Hx) = H
        Hy = p_tr(Hx, self.dims, self.sys)

        UxHxUx = self.Ux.conj().T @ Hx @ self.Ux
        UyHyUy = self.Uy.conj().T @ Hy @ self.Uy

        # Hessian products for quantum conditional entropy
        # D2Phi(X)[Hx] = Ux [log^[1](Dx) .* (Ux' Hx Ux)] Ux'
        #                - I ⊗ [Uy [log^[1](Dy) .* (Uy' pTr(Hx) Uy)] Uy']
        D2PhiH = self.Ux @ (self.D1x_log * UxHxUx) @ self.Ux.conj().T
        work = self.Uy @ (self.D1y_log * UyHyUy) @ self.Uy.conj().T
        D2PhiH -= i_kr(work, self.dims, self.sys)

        # ======================================================================
        # Hessian products with respect to t
        # ======================================================================
        # D2_t F(t, X)[Ht, Hx] = (Ht - D_X Phi(X)[Hx]) / z^2
        out[0][:] = (Ht - inp(self.DPhi, Hx)) * self.zi2

        # ======================================================================
        # Hessian products with respect to X
        # ======================================================================
        # D2_X F(t, X)[Ht, Hx] = -D2_t F(t, X)[Ht, Hx] * DPhi(X)
        #                        + D2Phi(X)[Hx] / z + X^-1 Hx X^-1
        out_X = -out[0] * self.DPhi
        out_X += self.zi * D2PhiH
        out_X += self.inv_X @ Hx @ self.inv_X
        out[1][:] = (out_X + out_X.conj().T) * 0.5

        return out

    def hess_congr(self, A):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()
        if not self.congr_aux_updated:
            self.congr_aux(A)

        p = A.shape[0]
        lhs = np.zeros((p, sum(self.dim)))

        Work0, Work1 = self.Work0, self.Work1
        Work2, Work3 = self.Work2, self.Work3

        # ======================================================================
        # Hessian products with respect to t
        # ======================================================================
        # D2_t F(t, X)[Ht, Hx] = (Ht - D_X Phi(X)[Hx]) / z^2
        DPhi_vec = self.DPhi.view(np.float64).reshape((-1, 1))
        out_t = self.At - (self.Ax_vec @ DPhi_vec).ravel()
        out_t *= self.zi2
        lhs[:, 0] = out_t

        # ======================================================================
        # Hessian products with respect to X
        # ======================================================================
        # Hessian products for quantum conditional entropy
        # D2Phi(X)[Hx] = Ux [log^[1](Dx) .* (Ux' Hx Ux)] Ux'
        #                - I ⊗ [Uy [log^[1](Dy) .* (Uy' pTr(Hx) Uy)] Uy']
        # Compute first term, i.e., Ux [log^[1](Dx) .* (Ux' Hx Ux)] Ux'
        congr_multi(Work2, self.Ux.conj().T, self.Ax, Work3)
        Work2 *= self.D1x_comb
        congr_multi(Work1, self.Ux, Work2, Work3)
        # Compute second term, i.e., I ⊗ [Uy [log^[1](Dy) .* (Uy' Hy Uy)] Uy']
        p_tr_multi(self.work1, self.Ax, self.dims, self.sys)  # Compute pTr(Ax)
        congr_multi(self.work2, self.Uy.conj().T, self.work1, self.work3)
        self.work2 *= self.D1y_log * self.zi
        congr_multi(self.work1, self.Uy, self.work2, self.work3)
        i_kr_multi(Work0, self.work1, self.dims, self.sys)
        # Subtract the two terms to obtain D2Phi(X)[Hx]
        Work1 -= Work0

        # Hessian product of barrier function
        # D2_X F(t, X)[Ht, Hx] = -D2_t F(t, X)[Ht, Hx] * DPhi(X)
        #                        + D2Phi(X)[Hx] / z + X^-1 Hx X^-1
        np.outer(out_t, self.DPhi, out=Work0.reshape(p, -1))
        Work1 -= Work0
        out_X = Work1.reshape((p, -1)).view(np.float64)
        lhs[:, 1:] = out_X

        # Multiply A (H A')
        return dense_dot_x(lhs, A.T)

    def invhess_prod_ip(self, out, H):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()
        if not self.invhess_aux_updated:
            self.update_invhessprod_aux()

        # Computes inverse Hessian product of the QCE barrier with a single vector (Ht, Hx)
        # See invhess_congr() for additional comments

        (Ht, Hx) = H
        Wx = Hx + Ht * self.DPhi

        # ======================================================================
        # Inverse Hessian products with respect to X
        # ======================================================================
        # X = [D2S(X)^-1 - D2S(X)^-1 pTr' N^-1 pTr D2S(X)^-1] Wx
        # Apply D2S(X)^-1 = (Ux ⊗ Ux) log^[1](Dx) (Ux' ⊗ Ux')
        UxWxUx = self.Ux.conj().T @ Wx @ self.Ux
        Hinv_x = self.Ux @ (self.D1x_comb_inv * UxWxUx) @ self.Ux.conj().T
        work = -p_tr(Hinv_x, self.dims, self.sys)

        # Solve linear system N \ ( ... )
        # Convert matrices to compact real vectors
        temp_vec = work.view(np.float64).reshape((-1, 1))
        temp_vec = self.F2C_op @ temp_vec
        # Solve system
        temp_vec = cho_solve(self.hess_schur_fact, temp_vec)
        # Expand compact real vectors back into full matrices
        temp_vec = self.F2C_op.T @ temp_vec
        work = temp_vec.T.view(self.dtype).reshape((self.n, self.n))

        # Apply pTr' = IKr
        temp = i_kr(work, self.dims, self.sys)
        # Apply D2S(X)^-1 = (Ux ⊗ Ux) log^[1](Dx) (Ux' ⊗ Ux')
        temp = self.Ux.conj().T @ temp @ self.Ux
        out_X = Hinv_x - self.Ux @ (self.D1x_comb_inv * temp) @ self.Ux.conj().T
        out[1][:] = (out_X + out_X.conj().T) * 0.5

        # ======================================================================
        # Inverse Hessian products with respect to t
        # ======================================================================
        # z^2 Ht + <DPhi(X), X>
        out[0][:] = Ht * self.z2 + inp(out_X, self.DPhi)

        return out

    def invhess_congr(self, A):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()
        if not self.invhess_aux_updated:
            self.update_invhessprod_aux()
        if not self.congr_aux_updated:
            self.congr_aux(A)

        # The inverse Hessian product applied on (Ht, Hx) for the QCE barrier is
        #     X = M \ Wx
        #     t = z^2 Ht + <DPhi(X), X>
        # where Wx = Hx + Ht DPhi(X)
        #     M = 1/z D2S(X) - 1/z pTr' D2S(pTr(X)) pTr + X^-1 ⊗ X^-1
        #       = (Ux ⊗ Ux) (1/z log + inv)^[1](Dx) (Ux' ⊗ Ux')
        #         - 1/z pTr' (Uy ⊗ Uy) log^[1](Dy) (Uy' ⊗ Uy') pTr
        #
        # Treating [pTr' D2S(pTr(X)) pTr] as a low-rank perturbation of D2S(X),
        # we can solve linear systems with M by using the matrix inversion lemma
        #     X = [D2S(X)^-1 - D2S(X)^-1 pTr' N^-1 pTr D2S(X)^-1] Wx
        # where
        #     N = 1/z (Uy ⊗ Uy) [log^[1](Dy)]^-1 (Uy' ⊗ Uy')
        #         - pTr (Ux ⊗ Ux) [(1/z log + inv)^[1](Dx)]^-1  (Ux' ⊗ Ux') pTr'

        p = A.shape[0]
        lhs = np.zeros((p, sum(self.dim)))

        Work0, Work1, work1 = self.Work0, self.Work1, self.work1
        Work2, Work3 = self.Work2, self.Work3

        # ======================================================================
        # Inverse Hessian products with respect to X
        # ======================================================================
        # X = [D2S(X)^-1 - D2S(X)^-1 pTr' N^-1 pTr D2S(X)^-1] Wx
        # Compute Wx = Hx + Ht DPhi(X)
        np.outer(self.At, self.DPhi, out=Work2.reshape((p, -1)))
        np.add(self.Ax, Work2, out=Work0)

        # Apply D2S(X)^-1 = (Ux ⊗ Ux) log^[1](Dx) (Ux' ⊗ Ux')
        congr_multi(Work2, self.Ux.conj().T, Work0, Work3)
        Work2 *= self.D1x_comb_inv
        congr_multi(Work0, self.Ux, Work2, Work3)
        # Apply pTr
        p_tr_multi(work1, Work0, self.dims, self.sys)
        work1 *= -1

        # Solve linear system N \ ( ... )
        # Convert matrices to compact real vectors
        work = work1.view(np.float64).reshape((p, -1)).T
        work = x_dot_dense(self.F2C_op, work)
        # Solve system
        work = cho_solve(self.hess_schur_fact, work)
        # Expand compact real vectors back into full matrices
        work = x_dot_dense(self.F2C_op.T, work)
        work1.view(np.float64).reshape((p, -1))[:] = work.T

        # Apply pTr' = iKr
        i_kr_multi(Work1, work1, self.dims, self.sys)
        # Apply D2S(X)^-1 = (Ux ⊗ Ux) log^[1](Dx) (Ux' ⊗ Ux')
        congr_multi(Work2, self.Ux.conj().T, Work1, Work3)
        Work2 *= self.D1x_comb_inv
        congr_multi(Work1, self.Ux, Work2, Work3)

        # Subtract previous expression from D2S(X)^-1 Wx to get X
        Work0 -= Work1
        out_X = Work0.reshape((p, -1)).view(np.float64)
        lhs[:, 1:] = out_X

        # ======================================================================
        # Inverse Hessian products with respect to t
        # ======================================================================
        # z^2 Ht + <DPhi(X), X>
        DPhi_vec = self.DPhi.view(np.float64).reshape((-1, 1))
        out_t = self.z2 * self.At
        out_t += (out_X @ DPhi_vec).ravel()
        lhs[:, 0] = out_t

        # Multiply A (H A')
        return dense_dot_x(lhs, A.T)

    def third_dir_deriv_axpy(self, out, H, a=True):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()
        if not self.dder3_aux_updated:
            self.update_dder3_aux()

        (Ht, Hx) = H
        Hy = p_tr(Hx, self.dims, self.sys)

        UxHxUx = self.Ux.conj().T @ Hx @ self.Ux
        UyHyUy = self.Uy.conj().T @ Hy @ self.Uy

        # Quantum conditional entropy Hessians
        D2PhiH = self.Ux @ (self.D1x_log * UxHxUx) @ self.Ux.conj().T
        work = self.Uy @ (self.D1y_log * UyHyUy) @ self.Uy.conj().T
        D2PhiH -= i_kr(work, self.dims, self.sys)

        # Quantum conditional entropy third order derivatives
        D3PhiHH = scnd_frechet(self.D2x_log, UxHxUx, UxHxUx, self.Ux)
        work = scnd_frechet(self.D2y_log, UyHyUy, UyHyUy, self.Uy)
        D3PhiHH -= i_kr(work, self.dims, self.sys)

        # Third derivative of barrier
        DPhiH = inp(self.DPhi, Hx)
        D2PhiHH = inp(D2PhiH, Hx)
        chi = Ht - DPhiH

        dder3_t = -2 * (self.zi**3) * (chi**2) - (self.zi**2) * D2PhiHH

        dder3_X = -dder3_t * self.DPhi
        dder3_X -= 2 * (self.zi**2) * chi * D2PhiH
        dder3_X += self.zi * D3PhiHH
        dder3_X -= 2 * self.inv_X @ Hx @ self.inv_X @ Hx @ self.inv_X
        dder3_X = (dder3_X + dder3_X.conj().T) * 0.5

        out[0][:] += dder3_t * a
        out[1][:] += dder3_X * a

        return out

    # ==========================================================================
    # Auxilliary functions
    # ==========================================================================
    def congr_aux(self, A):
        assert not self.congr_aux_updated

        iscomplex = self.iscomplex

        # Get slices and views of A matrix to be used in congruence computations
        if sp.sparse.issparse(A):
            A = A.tocsr()
        self.Ax_vec = A[:, 1:]

        if sp.sparse.issparse(A):
            A = A.toarray()
        Ax_dense = np.ascontiguousarray(A[:, 1:])
        self.At = A[:, 0]
        self.Ax = np.array([vec_to_mat(Ax_k, iscomplex) for Ax_k in Ax_dense])

        # Preallocate matrices we will need when performing these congruences
        self.Work0 = np.empty_like(self.Ax)
        self.Work1 = np.empty_like(self.Ax)
        self.Work2 = np.empty_like(self.Ax)
        self.Work3 = np.empty_like(self.Ax)

        p = A.shape[0]
        self.work1 = np.empty((p, self.n, self.n), dtype=self.dtype)
        self.work2 = np.empty((p, self.n, self.n), dtype=self.dtype)
        self.work3 = np.empty((p, self.n, self.n), dtype=self.dtype)

        self.congr_aux_updated = True

    def update_hessprod_aux(self):
        assert not self.hess_aux_updated
        assert self.grad_updated

        D1x_inv = np.reciprocal(np.outer(self.Dx, self.Dx))
        self.D1x_log = D1_log(self.Dx, self.log_Dx)
        self.D1x_comb = self.zi * self.D1x_log + D1x_inv

        self.D1y_log = D1_log(self.Dy, self.log_Dy)

        # Preparing other required variables
        self.zi2 = self.zi * self.zi

        self.hess_aux_updated = True

    def update_dder3_aux(self):
        assert not self.dder3_aux_updated
        assert self.hess_aux_updated

        self.D2x_log = D2_log(self.Dx, self.D1x_log)
        self.D2y_log = D2_log(self.Dy, self.D1y_log)

        self.dder3_aux_updated = True

    def update_invhessprod_aux(self):
        assert not self.invhess_aux_updated
        assert self.grad_updated
        assert self.hess_aux_updated
        if not self.invhess_aux_aux_updated:
            self.update_invhessprod_aux_aux()

        # Precompute and factorize the matrix
        #     N = z (Uy ⊗ Uy) [log^[1](Dy)]^-1 (Uy' ⊗ Uy')
        #         - pTr (Ux ⊗ Ux) [(1/z log + inv)^[1](Dx)]^-1  (Ux' ⊗ Ux') pTr'
        # which we will need to solve linear systems with the Hessian of our
        # barrier function

        self.z2 = self.z * self.z
        self.D1x_comb_inv = np.reciprocal(self.D1x_comb)

        work6, work7, work8 = self.work6, self.work7, self.work8
        Work6, Work7 = self.Work6, self.Work7
        Work8, Work9 = self.Work8, self.Work9

        # ======================================================================
        # Get first term, i.e., [z (Uy ⊗ Uy) [log^[1](Dy)]^-1 (Uy' ⊗ Uy')]
        # ======================================================================
        # Begin with (Uy' ⊗ Uy')
        congr_multi(work8, self.Uy.conj().T, self.E, work=work7)
        # Apply z [log^[1](Dy)]^-1
        work8 *= self.z * np.reciprocal(self.D1y_log)
        # Apply (Uy ⊗ Uy)
        congr_multi(work6, self.Uy, work8, work=work7)

        # ======================================================================
        # Get second term, i.e., [pTr (Ux ⊗ Ux) ...]
        # ======================================================================
        # Begin with [(Ux' ⊗ Ux') pTr']
        # Permute columns of Ux' so subsystems we are tracing out are in front
        temp = self.Ux.T.reshape(self.N, *self.dims)
        temp = np.transpose(temp, self.reordered_dims)
        # Obtain groups of columns of Ux' corresponding to
        #   Ux' (I ⊗ Eij) Ux = (Ux' [I ⊗ ei]) ([I ⊗ ej'] Ux)
        temp = temp.reshape(self.N, self.m, self.n)
        lhs = np.copy(temp.conj().transpose(2, 0, 1))
        rhs = np.copy(temp.transpose(2, 1, 0))
        # Use these groups of columns of Ux' to compute [(Ux' ⊗ Ux') pTr']
        # Compute the entries corresponding to Eii first
        np.matmul(lhs, rhs, out=Work9)
        Work8[self.diag_idxs] = Work9
        # Now compute entries corresponding to Eij for i =/= j
        rhs *= np.sqrt(0.5)
        t = 0
        for j in range(self.n):
            np.matmul(lhs[j], rhs[:j], out=Work9[:j])
            Work9_T = Work9[:j].conj().transpose(0, 2, 1)
            if self.iscomplex:
                # Get symmetric and skew-symmetric matrices corresponding to
                # real and complex basis vectors
                np.add(Work9[:j], Work9_T, out=Work8[t : t + 2 * j : 2])
                np.subtract(Work9[:j], Work9_T, out=Work8[t + 1 : t + 2 * j + 1 : 2])
                Work8[t + 1 : t + 2 * j + 1 : 2] *= -1j
                t += 2 * j + 1
            else:
                # Symmeterize matrix
                np.add(Work9[:j], Work9_T, out=Work8[t : t + j])
                t += j + 1

        # Apply [(1/z log + inv)^[1](Dx)]^-1/2
        Work8 *= self.D1x_comb_inv
        # Apply pTr (Ux ⊗ Ux)
        congr_multi(Work6, self.Ux, Work8, work=Work7)
        p_tr_multi(work7, Work6, self.dims, self.sys)

        # Subtract the two terms to obtain Schur complement then Cholesky factor
        work6 -= work7
        work = work6.view(np.float64).reshape((self.vn, -1))
        hess_schur = x_dot_dense(self.F2C_op, work.T)
        self.hess_schur_fact = cho_fact(hess_schur)

        self.invhess_aux_updated = True

    def update_invhessprod_aux_aux(self):
        assert not self.invhess_aux_aux_updated

        self.precompute_computational_basis()

        if self.iscomplex:
            self.diag_idxs = np.cumsum([i for i in range(3, 2 * self.n + 1, 2)])
            self.diag_idxs = np.append(0, self.diag_idxs)
        else:
            self.diag_idxs = np.cumsum([i for i in range(2, self.n + 1, 1)])
            self.diag_idxs = np.append(0, self.diag_idxs)

        not_sys = list(set(range(len(self.dims))) - set(self.sys))
        reordered_dims = self.sys + not_sys
        self.reordered_dims = [0] + [k + 1 for k in reordered_dims]

        self.F2C_op = get_full_to_compact_op(self.n, self.iscomplex)

        self.work6 = np.empty((self.vn, self.n, self.n), dtype=self.dtype)
        self.work7 = np.empty((self.vn, self.n, self.n), dtype=self.dtype)
        self.work8 = np.empty((self.vn, self.n, self.n), dtype=self.dtype)

        self.Work6 = np.empty((self.vn, self.N, self.N), dtype=self.dtype)
        self.Work7 = np.empty((self.vn, self.N, self.N), dtype=self.dtype)
        self.Work8 = np.empty((self.vn, self.N, self.N), dtype=self.dtype)
        self.Work9 = np.empty((self.n, self.N, self.N), dtype=self.dtype)

        self.invhess_aux_aux_updated = True



# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics

import math

import numpy as np


def vec_dim(side, iscomplex=False, compact=False):
    """Computes the dimension of a vectorized matrix.

    Parameters
    ----------
    side : :obj:`int`
        The dimension of the matrix.
    iscomplex : :obj:`bool`, optional
        Whether the matrix is Hermitian (``True``) or symmetric
        (``False``). The default is ``False``.
    compact : :obj:`bool`, optional
        Whether to assume a compact vector representation or not. The
        default is ``False``.

    Returns
    -------
    :obj:`int`
        The dimension of the vector.
    """
    if compact:
        if iscomplex:
            return side * side
        else:
            return side * (side + 1) // 2
    else:
        if iscomplex:
            return 2 * side * side
        else:
            return side * side


def mat_dim(len, iscomplex=False, compact=False):
    """Computes the dimension of the matrix correpsonding to a vector.

    Parameters
    ----------
    len : :obj:`int`
        The dimension of the vector.
    iscomplex : :obj:`bool`, optional
        Whether the matrix is Hermitian (``True``) or symmetric
        (``False``). The default is ``False``.
    compact : :obj:`bool`, optional
        Whether to assume a compact vector representation or not. The
        default is ``False``.

    Returns
    -------
    :obj:`int`
        The dimension of the matrix.
    """
    if compact:
        if iscomplex:
            return math.isqrt(len)
        else:
            return math.isqrt(1 + 8 * len) // 2
    else:
        if iscomplex:
            return math.isqrt(len // 2)
        else:
            return math.isqrt(len)


def mat_to_vec(mat, compact=False):
    r"""Reshapes a symmetric or Hermitian matrix into a column vector.

    If ``mat`` is of type :obj:`~numpy.float64` and ``compact=False``, then
    this performs the vectorization

    .. math::

        \begin{bmatrix}a & b & d \\ b & c & e \\ d & e & f\end{bmatrix}
        \mapsto
        \begin{bmatrix}a & b & d & b & c & e & d & e & f\end{bmatrix}^\top.

    If ``mat`` is of type :obj:`~numpy.float64` and ``compact=True``, then
    this performs the vectorization

    .. math::

        \begin{bmatrix}a & b & d \\ b & c & e \\ d & e & f\end{bmatrix}
        \mapsto\begin{bmatrix}
            a & \sqrt{2} b & c & \sqrt{2} d & \sqrt{2}e & f
        \end{bmatrix}^{\top}.

    If ``mat`` is of type :obj:`~numpy.complex128` and ``compact=False``,
    then this performs the vectorization

    .. math::

        \begin{bmatrix}
            a & b+ci & e+fi \\ b-ci & d & g+hi \\ e-fi & g-hi & k
        \end{bmatrix}\mapsto
        \begin{bmatrix}
            a & 0 & b & c & e & f & b & -c & d & 0
            & g & h & e & -f & g & -h & k & 0
        \end{bmatrix}^{\top}.

    If ``mat`` is of type :obj:`~numpy.complex128` and ``compact=True``,
    then this performs the vectorization

    .. math::

        \begin{bmatrix}
            a & b+ci & e+fi \\ b-ci & d & g+hi \\ e-fi & g-hi & k
        \end{bmatrix}\mapsto
        \begin{bmatrix}
            a & \sqrt{2} b & \sqrt{2} c & d & \sqrt{2} e & \sqrt{2} f &
            \sqrt{2} g & \sqrt{2} h & k
        \end{bmatrix}^{\top}.

    Parameters
    ----------
    mat : :class:`~numpy.ndarray`
        Input matrix to vectorize, either of type :obj:`~numpy.float64`
        or :obj:`~numpy.complex128`.
    compact : :obj:`bool`, optional
        Whether to convert to a compact vector representation or not.
        The default is ``False``.

    Returns
    -------
    :class:`~numpy.ndarray`
        The resulting vectorized matrix.
    """
    if np.isscalar(mat):
        mat = np.array([[mat]])
    
    iscomplex = np.iscomplexobj(mat)

    if iscomplex:
        mat = mat.astype(np.complex128, copy=False)
    else:
        mat = mat.astype(np.float64, copy=False)

    n = mat.shape[0]
    vn = vec_dim(n, iscomplex=iscomplex, compact=compact)

    if compact:
        rt2 = np.sqrt(2.0)
        vec = np.empty((vn, 1))

        k = 0
        for j in range(n):
            for i in range(j):
                vec[k] = mat[i, j].real * rt2
                k += 1
                if iscomplex:
                    vec[k] = mat[i, j].imag * rt2
                    k += 1
            vec[k] = mat[j, j].real
            k += 1

        return vec
    else:
        mat = np.ascontiguousarray(mat)
        return mat.view(np.float64).reshape(-1, 1).copy()


def vec_to_mat(vec, iscomplex=False, compact=False):
    r"""Reshapes a column vector into a symmetric or Hermitian matrix.

    If ``iscomplex=False`` and ``compact=False``, then this returns the
    matrix

    .. math::

        \begin{bmatrix}a & b & d & b & c & e & d & e & f\end{bmatrix}^\top
        \mapsto
        \begin{bmatrix}a & b & d \\ b & c & e \\ d & e & f\end{bmatrix}.

    If ``iscomplex=False`` and ``compact=True``, then this performs the
    matrix

    .. math::

        \begin{bmatrix}
            a & \sqrt{2} b & c & \sqrt{2} d & \sqrt{2}e & f
        \end{bmatrix}^{\top}\mapsto
        \begin{bmatrix}a & b & d \\ b & c & e \\ d & e & f\end{bmatrix}.

    If ``iscomplex=True`` and ``compact=False``, then this returns the
    matrix

    .. math::

        \begin{bmatrix}
            a & 0 & b & c & e & f & b & -c & d & 0
            & g & h & e & -f & g & -h & k & 0
        \end{bmatrix}^{\top}\mapsto
        \begin{bmatrix}
            a & b+ci & e+fi \\ b-ci & d & g+hi \\ e-fi & g-hi & k
        \end{bmatrix}.

    If ``iscomplex=True`` and ``compact=True``, then this returns the
    matrix

    .. math::

        \begin{bmatrix}
            a & \sqrt{2} b & \sqrt{2} c & d & \sqrt{2} e & \sqrt{2} f &
            \sqrt{2} g & \sqrt{2} h & k
        \end{bmatrix}^{\top}\mapsto
        \begin{bmatrix}
            a & b+ci & e+fi \\ b-ci & d & g+hi \\ e-fi & g-hi & k
        \end{bmatrix}.


    Parameters
    ----------
    mat : :class:`~numpy.ndarray`
        Input vector to reshape into a matrix.
    iscomplex : :obj:`bool`, optional
        Whether the resulting matrix is Hermitian (``True``) or symmetric
        (``False``). The default is ``False``.
    compact : :obj:`bool`, optional
        Whether to convert from a compact vector representation or not.
        The default is ``False``.

    Returns
    -------
    :class:`~numpy.ndarray`
        The resulting vector.
    """
    vn = vec.size
    n = mat_dim(vn, iscomplex=iscomplex, compact=compact)
    dtype = np.complex128 if iscomplex else np.float64

    if compact:
        irt2 = np.sqrt(0.5)
        mat = np.empty((n, n), dtype=dtype)

        k = 0
        for j in range(n):
            for i in range(j):
                if iscomplex:
                    mat[i, j] = (vec[k, 0] + vec[k + 1, 0] * 1j) * irt2
                    k += 2
                else:
                    mat[i, j] = vec[k, 0] * irt2
                    k += 1
                mat[j, i] = mat[i, j].conjugate()

            mat[j, j] = vec[k, 0]
            k += 1

        return mat
    else:
        if iscomplex:
            n = math.isqrt(vn // 2)
            mat = vec.reshape((-1, 2)).view(np.complex128).reshape(n, n)
            return (mat + mat.conj().T) * 0.5
        else:
            n = math.isqrt(vn)
            mat = vec.reshape((n, n))
            return (mat + mat.T) * 0.5


def lin_to_mat(lin, dims, iscomplex=False, compact=(False, True)):
    """Computes the matrix corresponding to a linear map from vectorized
    symmetric matrices to vectorized symmetric matrices.

    Parameters
    ----------
    lin : :obj:`callable`
        Linear operator sending symmetric matrices to symmetric matrices.
    dims : :obj:`tuple` of :obj:`int`
        The dimensions ``(ni, no)`` of the input and output matrices of the
        linear operator.
    iscomplex : :obj:`bool`, optional
        Whether the matrix to vectorize is Hermitian (``True``) or
        symmetric (``False``). Default is ``False``.
    compact : :obj:`tuple` of :obj:`bool`, optional
        Whether to use a compact vector representation or not for the input
        and output matrices, respectively. Default is ``(False, True)``.

    Returns
    -------
    :class:`~numpy.ndarray`
        The matrix representation of the given linear operator.
    """
    vni = vec_dim(dims[0], iscomplex=iscomplex, compact=compact[0])
    vno = vec_dim(dims[1], iscomplex=iscomplex, compact=compact[1])
    mat = np.zeros((vno, vni))

    for k in range(vni):
        H = np.zeros((vni, 1))
        H[k] = 1.0
        H_mat = vec_to_mat(H, iscomplex=iscomplex, compact=compact[0])
        lin_H = lin(H_mat)
        vec_out = mat_to_vec(lin_H, compact=compact[1])
        mat[:, [k]] = vec_out

    return mat


def eye(n, iscomplex=False, compact=(False, True)):
    """Computes the matrix representation of the identity map for
    vectorized symmetric or Hermitian matrices.

    Parameters
    ----------
    n : :obj:`int`
        The dimensions of the ``(n, n)`` matrix the identity is acting on.
    iscomplex : :obj:`bool`, optional
        Whether the matrix to vectorize is Hermitian (``True``) or
        symmetric (``False``). Default is ``False``.
    compact : :obj:`tuple` of :obj:`bool`, optional
        Whether to use a compact vector representation or not for the input
        and output matrices, respectively. Default is ``(False, True)``.

    Returns
    -------
    :class:`~numpy.ndarray`
        The matrix representation of the identity superoperator.
    """
    return lin_to_mat(lambda X: X, (n, n), iscomplex=iscomplex, compact=compact)


def get_full_to_compact_op(n, iscomplex=False):
    import scipy

    dim_compact = n * n if iscomplex else n * (n + 1) // 2
    dim_full = 2 * n * n if iscomplex else n * n

    rows = np.zeros(dim_full)
    cols = np.zeros(dim_full)
    vals = np.zeros(dim_full)

    irt2 = np.sqrt(0.5)

    row = 0
    k = 0
    for j in range(n):
        for i in range(j):
            rows[k : k + 2] = row
            cols[k : k + 2] = (
                [2 * (i + j * n), 2 * (j + i * n)]
                if iscomplex
                else [i + j * n, j + i * n]
            )
            vals[k : k + 2] = irt2
            k += 2
            row += 1

            if iscomplex:
                rows[k : k + 2] = row
                cols[k : k + 2] = [2 * (i + j * n) + 1, 2 * (j + i * n) + 1]
                vals[k : k + 2] = [-irt2, irt2]
                k += 2
                row += 1

        rows[k] = row
        cols[k] = 2 * j * (n + 1) if iscomplex else j * (n + 1)
        vals[k] = 1.0
        k += 1
        row += 1

    shape = (dim_compact, dim_full)
    return scipy.sparse.coo_matrix((vals, (rows, cols)), shape=shape)



# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics

import numpy as np
import scipy as sp

from qics._utils.linalg import abs_max, is_full_col_rank, scale_axis


class Model:
    r"""A class representing an instance of the standard form primal

    .. math::

        \min_{x \in \mathbb{R}^n} &&& c^\top x

        \text{s.t.} &&& b - Ax = 0

         &&& h - Gx \in \mathcal{K},

    and dual

    .. math::

        \max_{y \in \mathbb{R}^p, z \in \mathbb{R}^q} &&&
        -b^\top y - h^\top z

        \text{s.t.} &&& c + A^\top y + G^\top z = 0

         &&& z \in \mathcal{K}_*,

    conic programs, where :math:`c\in\mathbb{R}^n`,
    :math:`b\in\mathbb{R}^p`, :math:`h\in\mathbb{R}^q`,
    :math:`A\in\mathbb{R}^{p\times n}`, :math:`G\in\mathbb{R}^{q\times n}`,
    and :math:`\mathcal{K}\subset\mathbb{R}^{q}` is a convex, proper cone
    with dual cone :math:`\mathcal{K}_*\subset\mathbb{R}^{q}`.


    Parameters
    ----------
    c : :class:`~numpy.ndarray`
        2D :obj:`~numpy.float64` array of size ``(n, 1)`` representing the
        linear objective :math:`c`.
    A : :class:`~numpy.ndarray` or :class:`~scipy.sparse.sparray`, optional
        2D :obj:`~numpy.float64` array of size ``(p, 1)`` representing
        linear equality constraint matrix :math:`A`. The default is
        ``numpy.empty((0, n))``, i.e., there are no linear equalitiy
        constraints.
    b : :class:`~numpy.ndarray`, optional
        2D :obj:`~numpy.float64` array of size ``(p, 1)`` representing
        linear equality constraint vector :math:`b`. The default is
        ``numpy.zeros((p, 1))``, i.e., :math:`b=0`.
    G : :class:`~numpy.ndarray` or :class:`~scipy.sparse.sparray`, optional
        2D :obj:`~numpy.float64` array of size ``(q, n)`` representing
        linear cone constraint matrix :math:`G`. The default is
        ``-scipy.sparse.eye(n)``, i.e., cone constraints are of the
        simplified form :math:`x+h\in\mathcal{K}`.
    h : :class:`~numpy.ndarray`, optional
        2D :obj:`~numpy.float64` array of size ``(q, 1)`` representing
        linear cone constraint vector :math:`h`. The default is
        ``numpy.zeros((q, 1))``, i.e., :math:`h=0`.
    cones : :class:`list` of :mod:`~qics.cones`, optional
        Cartesian product of cones :math:`\mathcal{K}`. Default is ``[]``
        i.e., there are no conic constraints.
    offset : :class:`float`, optional
        Constant offset term to add to the objective function. Default is
        ``0``.
    """

    def __init__(self, c, A=None, b=None, G=None, h=None, cones=None, offset=0.0):
        # Intiialize model parameters and default values for missing data
        self.n_orig = self.n = np.size(c)
        self.p_orig = self.p = np.size(b) if (b is not None) else 0
        self.q_orig = self.q = np.size(h) if (h is not None) else self.n

        # Make copies of everything so we don't overwrite data matrices
        self.c = c.copy()
        self.A = A.copy() if (A is not None) else np.empty((0, self.n))
        self.b = b.copy() if (b is not None) else np.empty((0, 1))
        self.G = G.copy() if (G is not None) else -sp.sparse.eye(self.n).tocsr()
        self.h = h.copy() if (h is not None) else np.zeros((self.n, 1))
        self.cones = cones
        self.offset = offset

        # Make sure dimensions of matrices all make sense
        if self.c.shape[1] != 1 or self.b.shape[1] != 1 or self.h.shape[1] != 1:
            raise ValueError(
                "The arrays c, b, and h should all have a dimension (-1, 1), but have "
                "dimensions", self.c.shape, ",", self.b.shape, ", and", self.h.shape
            )  # fmt: skip
        if self.A.shape[1] != self.n or self.G.shape[1] != self.n:
            raise ValueError(
                "The length of c should match the number of columns in A and G, but "
                "have dimensions", self.c.shape, ",", self.A.shape, ", and", 
                self.G.shape
            )  # fmt: skip
        if self.A.shape[0] != self.p:
            raise ValueError(
                "The length of b should match the number of rows in A, but have "
                "dimensions", self.b.shape, "and", self.A.shape
            )  # fmt: skip
        if self.G.shape[0] != self.q:
            raise ValueError(
                "The length of h should match the number of rows in G, but have "
                "dimensions", self.b.shape, "and", self.A.shape
            )  # fmt: skip

        # Barrier parameter
        self.nu = 1 + sum([cone.nu for cone in cones])

        # Get properties of the problem
        self.issymmetric = all([cone.get_issymmetric() for cone in self.cones])
        self.iscomplex = any([cone.get_iscomplex() for cone in self.cones])

        # Check if model uses A or G matrices
        self.use_G = not _is_like_eye(self.G)
        self.use_A = (A is not None) and (np.prod(A.shape) > 0)

    def _preprocess(self, use_invhess=False, init_pnt=None):
        SPARSE_THRESHOLD = 0.01
        cone_idxs = self.cone_idxs = _build_cone_idxs(self.q, self.cones)

        # Sparsify A and G if they are sufficiently sparse
        self.A = _sparsify(self.A, SPARSE_THRESHOLD, "csr")
        if self.use_G:
            self.G = _sparsify(self.G, SPARSE_THRESHOLD, "csr")

        # Restructure to allow for avoiding inverse Hessian oracles
        if self.use_G and not use_invhess:
            self._restructure(init_pnt)

        # Rescale model
        self._rescale()

        # Precompute transposes of A and G for faster sparse operations
        self.A_T = self.A.T.tocsr() if sp.sparse.issparse(self.A) else self.A.T
        self.G_T = self.G.T.tocsr() if sp.sparse.issparse(self.G) else self.G.T

        # Get slices of A or G matrices correpsonding to each cone
        # and some other handy precomputations
        if self.use_G:
            self.G_T_views = [self.G_T[:, idxs_k] for idxs_k in cone_idxs]
            self.G_T_views = _sparsify(self.G_T_views, SPARSE_THRESHOLD)

            # Need a dense A' to do Cholesky solves on
            if sp.sparse.issparse(self.A):
                self.A_T_dense = self.A_T.toarray()
            else:
                self.A_T_dense = self.A_T

            issparse_list = [sp.sparse.issparse(Gk) for Gk in self.G_T_views]
            self.issparse = any(issparse_list)

        elif self.use_A:
            # After rescaling, G is an easily invertible square diagonal matrix
            self.G_inv = np.reciprocal(self.G.diagonal()).reshape((-1, 1))

            self.A_invG = scale_axis(self.A.copy(), scale_cols=self.G_inv)
            if sp.sparse.issparse(self.A_invG):
                self.A_invG = self.A_invG.tocsr()

            self.A_invG_views = [self.A_invG[:, idxs_k] for idxs_k in cone_idxs]
            self.A_invG_views = _sparsify(self.A_invG_views, SPARSE_THRESHOLD)

            issparse_list = [sp.sparse.issparse(Ak) for Ak in self.A_invG_views]
            self.issparse = any(issparse_list)

        else:
            self.G_inv = np.reciprocal(self.G.diagonal()).reshape((-1, 1))
            self.issparse = True

        return

    def _restructure(self, init_pnt=None):
        # Restructures the conic program into
        #     min  <c,x1>
        #     s.t  A*x1 = b,  x2 = 0,  x3 = 1
        #          h1*x2 + h2*x3 - G*x1 ∈ K
        # where h1 is an interior point of K and h2 = h. This allows us to
        # solve problems using the cone K'={x : G*x ∈ K}.
        from qics.point import VecProduct

        n = self.n
        self.x_offset = np.zeros((n, 1))

        # Add variable x2 and constraint x2 = 0 if necessary
        # Find an interior point of K and normalize it
        if init_pnt is None:
            s_init = VecProduct(self.cones)
            s_init.vec.fill(np.nan)
        else:
            # If user gives us an inital s, then use this instead
            s_init = init_pnt.s

        for k, cone_k in enumerate(self.cones):
            if any(np.isnan(s_init.vecs[k])):
                cone_k.get_init_point(s_init[k])
        s_norm = np.sum(np.abs(s_init.vec))

        G_temp = _hstack((self.G, -s_init.vec / s_norm))
        if is_full_col_rank(G_temp):
            A_new_col = sp.sparse.coo_matrix((self.p, 1))
            A_new_row = sp.sparse.coo_matrix(([1.0], ([0], [n])), (1, n + 1))
            self.c = np.vstack((self.c, np.array([[0.0]])))
            self.A = _vstack((_hstack((self.A, A_new_col)), A_new_row))
            self.b = np.vstack((self.b, np.array([[0.0]])))
            self.G = G_temp

            self.x_offset = np.vstack((self.x_offset, np.array([[0.0]])))

            self.use_A = True
            (n, _) = (self.n, self.p) = (self.n + 1, self.p + 1)

        # Add variable x3 and constraint x3 = 1 if necessary
        if np.any(self.h):
            h_norm = np.sum(np.abs(self.h))
            G_temp = _hstack((self.G, -self.h / h_norm))
            if is_full_col_rank(G_temp):
                A_new_col = sp.sparse.coo_matrix((self.p, 1))
                A_new_row = sp.sparse.coo_matrix(([1.0], ([0], [n])), (1, n + 1))
                self.c = np.vstack((self.c, np.array([[0.0]])))
                self.A = _vstack((_hstack((self.A, A_new_col)), A_new_row))
                self.b = np.vstack((self.b, np.array([[h_norm]])))
                self.G = G_temp
                self.h = np.zeros((self.q, 1))

                self.x_offset = np.vstack((self.x_offset, np.array([[0.0]])))

                self.use_A = True
                (n, _) = (self.n, self.p) = (self.n + 1, self.p + 1)
            else:
                self.x_offset = sp.sparse.linalg.lsqr(self.G, self.h)[0]
                self.x_offset = self.x_offset.reshape((-1, 1))
                self.offset += (self.c.T @ self.x_offset)[0, 0]
                self.b = self.b - self.A @ self.x_offset
                self.h = np.zeros((self.q, 1))

        if self.x_offset is None:
            self.x_offset = np.zeros((n, 1))

    def _rescale(self):
        # Rescale c
        self.c_scale = np.maximum.reduce([np.abs(self.c.ravel()), 
                                          abs_max(self.A, axis=0), 
                                          abs_max(self.G, axis=0)])  # fmt: skip
        self.c_scale = np.sqrt(self.c_scale).reshape((-1, 1))

        # Rescale b
        self.b_scale = np.maximum.reduce([np.abs(self.b.ravel()), 
                                          abs_max(self.A, axis=1)])  # fmt: skip
        self.b_scale = np.sqrt(self.b_scale).reshape((-1, 1))

        # Rescale h
        # Note we can only scale each cone by a positive factor, and
        # we can't scale each individual variable by a different factor
        # (except for the nonnegative orthant)
        from qics.cones import NonNegOrthant

        self.h_scale = np.zeros((self.q, 1))
        h_absmax = np.abs(self.h.ravel())
        G_absmax = abs_max(self.G, axis=1)
        for k, cone_k in enumerate(self.cones):
            idxs = self.cone_idxs[k]
            if isinstance(cone_k, NonNegOrthant):
                self.h_scale[idxs, 0] = np.maximum.reduce([h_absmax[idxs], 
                                                           G_absmax[idxs]])  # fmt: skip
            else:
                self.h_scale[idxs, 0] = np.max([h_absmax[idxs], G_absmax[idxs]])
        self.h_scale = np.sqrt(self.h_scale)

        # Ensure there are no divide by zeros
        EPS = np.finfo(self.b_scale.dtype).eps
        self.c_scale[self.c_scale < EPS] = 1.0
        self.b_scale[self.b_scale < EPS] = 1.0
        self.h_scale[self.h_scale < EPS] = 1.0

        # Rescale data
        self.c /= self.c_scale
        self.b /= self.b_scale
        self.h /= self.h_scale
        self.A = scale_axis(self.A,
                            scale_rows=np.reciprocal(self.b_scale),
                            scale_cols=np.reciprocal(self.c_scale))  # fmt: skip
        self.G = scale_axis(self.G,
                            scale_rows=np.reciprocal(self.h_scale),
                            scale_cols=np.reciprocal(self.c_scale))  # fmt: skip

        return


def _is_like_eye(A, tol=1e-10):
    if A.shape[0] != A.shape[1]:
        return False
    n = A.shape[0]
    if sp.sparse.issparse(A):
        A_minus_eye = sp.sparse.eye(n) - abs(A)
        return sp.sparse.linalg.norm(A_minus_eye) < tol
    else:
        A_minus_eye = np.eye(n) - abs(A)
        return np.linalg.norm(A_minus_eye) < tol


def _vstack(tup):
    if isinstance(tup, tuple):
        tup = list(tup)

    if sp.sparse.issparse(tup[0]):
        for k in range(1, len(tup)):
            tup[k] = sp.sparse.coo_matrix(tup[k])
        return sp.sparse.vstack(tup)
    else:
        for k in range(1, len(tup)):
            if sp.sparse.issparse(tup[k]):
                tup[k] = tup[k].toarray()
        return np.vstack(tup)


def _hstack(tup):
    if isinstance(tup, tuple):
        tup = list(tup)

    if sp.sparse.issparse(tup[0]):
        for k in range(1, len(tup)):
            tup[k] = sp.sparse.coo_matrix(tup[k])
        return sp.sparse.hstack(tup)
    else:
        for k in range(1, len(tup)):
            if sp.sparse.issparse(tup[k]):
                tup[k] = tup[k].toarray()
        return np.hstack(tup)


def _build_cone_idxs(q, cones):
    cone_idxs = []
    prev_idx = 0
    for cone in cones:
        dim_k = np.sum(cone.dim)
        cone_idxs.append(slice(prev_idx, prev_idx + dim_k))
        prev_idx += dim_k

    if prev_idx != q:
        raise ValueError(
            "The dimensions of h and the cone K should be match, but are instead have ",
            "the dimensions", q, "and", prev_idx
        )  # fmt: skip

    return cone_idxs


def _sparsify(A, threshold, format="coo"):
    def sparsify_single(A, threshold, format):
        if A.size == 0:
            if sp.sparse.issparse(A):
                A = A.toarray()
            return A

        if sp.sparse.issparse(A):
            if A.nnz / np.prod(A.shape) > threshold:
                return A.toarray()
            else:
                if format == "coo":
                    return A.tocoo()
                elif format == "csr":
                    return A.tocsr()
        else:
            if np.count_nonzero(A) / A.size < threshold:
                if format == "coo":
                    return sp.sparse.coo_matrix(A)
                elif format == "csr":
                    return sp.sparse.csr_matrix(A)
        return A

    if isinstance(A, list):
        return [sparsify_single(A_k, threshold, format) for A_k in A]
    else:
        return sparsify_single(A, threshold, format)



# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics

import numpy as np

from qics._utils.linalg import inp


class Cone:
    """Base class for cones"""

    def __init__(self):
        pass

    def get_issymmetric(self):
        return False

    def get_iscomplex(self):
        return False

    def zeros(self):
        out = []
        for dim_k, type_k in zip(self.dim, self.type):
            if type_k == "r":
                out += [np.zeros((dim_k, 1))]
            elif type_k == "s":
                n_k = int(np.sqrt(dim_k))
                out += [np.zeros((n_k, n_k))]
            elif type_k == "h":
                n_k = int(np.sqrt(dim_k // 2))
                out += [np.zeros((n_k, n_k), dtype=np.complex128)]
        return out

    def prox(self):
        assert self.feas_updated
        if not self.grad_updated:
            self.update_grad()

        # Proximity measure is given by <psi, H^-1 psi>
        # where psi = z/mu + g(s)
        psi = [dual_k + grad_k for (dual_k, grad_k) in zip(self.dual, self.grad)]

        H_psi = self.zeros()
        self.invhess_prod_ip(H_psi, psi)

        return sum([inp(H_psi_k, psi_k) for (H_psi_k, psi_k) in zip(H_psi, psi)])

    def set_point(self, primal, dual=None, a=True):
        self.primal = [primal_k * a for primal_k in primal]
        if dual is not None:
            self.dual = [dual_k * a for dual_k in dual]
        else:
            self.dual = None

        self.feas_updated = False
        self.grad_updated = False
        self.hess_aux_updated = False
        self.invhess_aux_updated = False
        self.dder3_aux_updated = False
        self.nt_aux_updated = False

    def set_dual(self, dual, a=True):
        self.dual = [dual_k * a for dual_k in dual]

    def get_dual_feas(self):
        return True

    def grad_ip(self, out):
        assert self.feas_updated
        if not self.grad_updated:
            self.update_grad()

        for out_k, grad_k in zip(out, self.grad):
            out_k[:] = grad_k

        return out

    def precompute_computational_basis(self, n=None):
        if hasattr(self, "E"):
            return

        n = self.n if (n is None) else n
        vn = n * n if self.iscomplex else n * (n + 1) // 2

        # Computational basis for symmetric/Hermitian matrices
        rt2 = np.sqrt(0.5)
        self.E = np.zeros((vn, n, n), dtype=self.dtype)
        k = 0
        for j in range(n):
            for i in range(j):
                self.E[k, i, j] = rt2
                self.E[k, j, i] = rt2
                k += 1
                if self.iscomplex:
                    self.E[k, i, j] = rt2 * 1j
                    self.E[k, j, i] = rt2 * -1j
                    k += 1
            self.E[k, j, j] = 1.0
            k += 1

    # Functions that the child class has to implement
    def get_init_point(self, out):
        """Returns a central primal-dual point (s0, z0) satisfying

            z0 = -F'(s0)

        and stores this point in-place in out.

        Parameters
        ----------
        out : ndarray or list of ndarray
            Preallocated vector to store the central point in place.
        """
        pass

    def get_feas(self):
        """Returns whether current primal point s is in the interior of the cone K.

        Returns
        ----------
        bool
            Whether current primal point s is in the interior of the cone K.
        """
        pass

    def update_grad(self):
        """Compute the gradient F'(s) of the barrier function and store in self.grad."""
        pass

    def hess_prod_ip(self, out, H):
        """Compute the Hessian product D2F(s)[H] of the barrier function in the
        direction of H, and store this in-place in out.

        Parameters
        ----------
        out : ndarray or list of ndarray
            Preallocated vector to store the Hessian product in place.
        H : ndarray or list of ndarray
            The direction to compute the second derivative in.
        """
        pass

    def hess_congr(self, A):
        """Compute the congruence transform AH(s)A' with the Hessian matrix H of
        the barrier function.

        Parameters
        ----------
        A : ndarray
            Matrix of size (p, n), where n is the dimension of the cone K.
            Should correspond to a linear constraint matrix from the conic
            program definition.

        Returns
        ----------
        ndarray
            The matrix product AH(s)A'.
        """
        pass

    def invhess_prod_ip(self, out, H):
        """Compute the inverse Hessian product D2F(s)^-1[H] of the barrier
        function in the direction of H, and store this in-place in out.

        Parameters
        ----------
        out : ndarray or list of ndarray
            Preallocated vector to store the inverse Hessian product in place.
        H : ndarray or list of ndarray
            The direction to compute the inverse second derivative in.
        """
        pass

    def invhess_congr(self, A):
        """Compute the congruence transform A(H(s)^-1)A' with the inverse Hessian
        matrix H(s)^-1 of the barrier function.

        Parameters
        ----------
        A : ndarray
            Matrix of size (p, n), where n is the dimension of the cone K.
            Should correspond to a linear constraint matrix from the conic
            program definition.

        Returns
        ----------
        ndarray
            The matrix product A(H(s)^-1)A'.
        """
        pass

    def third_dir_deriv_axpy(self, out, H, a=True):
        """Compute the third directional derivative in direction H and add
        it to an existing vector, i.e.,

            out <-- out + a * D3F(s)[H, H]

        Parameters
        ----------
        out : ndarray or list of ndarray
            Vector to add the third directional derivative to.
        H : ndarray or list of ndarray
            The direction to compute the third directional derivative in.
        a : float, optional
            Amount to scale the third directional derivative by. Default is 1.
        """
        pass


class SymCone(Cone):
    """Base class for symmetric cones. These cones have a NT scaling
    point w and scaled variable lambda such that

        H(w) s = z  <==> lambda := W^-T s = W z

    where H(w) = W^T W.
    """

    def get_issymmetric(self):
        return True

    # Functions that the child class has to implement
    def nt_prod_ip(self, out, H):
        """Compute the Hessian product D2F(w)[H] of the barrier function in the
        direction of H, and store this in-place in out.

        Parameters
        ----------
        out : ndarray or list of ndarray
            Preallocated vector to store the Hessian product in place.
        H : ndarray or list of ndarray
            The direction to compute the second derivative in.
        """
        pass

    def nt_congr(self, A):
        """Compute the congruence transform AH(w)A' with the Hessian matrix H of the barrier function.

        Parameters
        ----------
        A : ndarray
            Matrix of size (p, n), where n is the dimension of the cone K.
            Should correspond to a linear constraint matrix from the conic
            program definition.

        Returns
        ----------
        ndarray
            The matrix product AH(w)A'.
        """
        pass

    def invnt_prod_ip(self, out, H):
        """Compute the inverse Hessian product D2F(w)^-1[H] of the barrier function in the
        direction of H, and store this in-place in out.

        Parameters
        ----------
        out : ndarray or list of ndarray
            Preallocated vector to store the inverse Hessian product in place.
        H : ndarray or list of ndarray
            The direction to compute the inverse second derivative in.
        """
        pass

    def invnt_congr(self, A):
        """Compute the congruence transform A(H(w)^-1)A' with the inverse Hessian
        matrix H(w)^-1 of the barrier function.

        Parameters
        ----------
        A : ndarray
            Matrix of size (p, n), where n is the dimension of the cone K.
            Should correspond to a linear constraint matrix from the conic
            program definition.

        Returns
        ----------
        ndarray
            The matrix product A(H(w)^-1)A'.
        """
        pass

    def comb_dir(self, out, ds, dz, sigma_mu):
        r"""Compute the residual rs where rs is given as the lhs of

            Lambda o (W dz + W^-T ds) = -Lambda o Lambda - (W^-T ds_a) o (W dz_a)
                                        + sigma * mu * I

        which is rearranged into the form H ds + dz = rs, i.e.,

            rs := W^-1 [ Lambda \ (-Lambda o Lambda - (W^-T ds_a) o (W dz_a) + sigma*mu I) ]

        See: [Section 5.4]https://www.seas.ucla.edu/~vandenbe/publications/coneprog.pdf

        Parameters
        ----------
        out : ndarray or list of ndarray
            Preallocated vector to store the residual in place.
        ds : ndarray or list of ndarray
            Vector representing the primal step direction.
        dz : ndarray or list of ndarray
            Vector representing the dual step direction.
        sigma_mu : float
            Value of the product simga*mu.
        """
        pass

    def step_to_boundary(self, ds, dz):
        """Compute the maximum step alpha in [0, 1] we can take such that

            s + alpha ds >= 0
            z + alpha dz >= 0

        See: [Section 8.3]https://www.seas.ucla.edu/~vandenbe/publications/coneprog.pdf

        Parameters
        ----------
        ds : ndarray or list of ndarray
            Vector representing the primal step direction.
        dz : ndarray or list of ndarray
            Vector representing the dual step direction.
        """
        pass


def get_central_ray_relentr(x_dim):
    if x_dim <= 10:
        return central_rays_relentr[x_dim - 1, :]

    # use nonlinear fit for higher dimensions
    rtx_dim = np.sqrt(x_dim)
    if x_dim <= 20:
        t = 1.2023 / rtx_dim - 0.0150
        x = -0.3057 / rtx_dim + 0.9720
        y = 0.4320 / rtx_dim + 1.0125
    else:
        t = 1.1513 / rtx_dim - 0.0069
        x = -0.4247 / rtx_dim + 0.9961
        y = 0.4873 / rtx_dim + 1.0008

    return [t, x, y]


central_rays_relentr = np.array(
    [
        [0.827838399, 0.805102002, 1.290927710],
        [0.708612491, 0.818070436, 1.256859152],
        [0.622618845, 0.829317078, 1.231401007],
        [0.558111266, 0.838978355, 1.211710886],
        [0.508038610, 0.847300430, 1.196018952],
        [0.468039614, 0.854521306, 1.183194752],
        [0.435316653, 0.860840990, 1.172492396],
        [0.408009282, 0.866420016, 1.163403373],
        [0.384838611, 0.871385497, 1.155570329],
        [0.364899121, 0.875838067, 1.148735192],
    ]
)


def get_central_ray_entr(x_dim):
    if x_dim <= 10:
        return central_rays_entr[x_dim - 1, :]

    # use nonlinear fit for higher dimensions
    t = np.exp(-0.9985 * np.log(x_dim) + 0.6255)
    u = 1.6990 - np.exp(-1.0556 * np.log(x_dim) - 0.9553)
    x = 1.0000 - np.exp(-0.9919 * np.log(x_dim) - 0.8646)

    return [t, u, x]


central_rays_entr = np.array(
    [
        [0.827838399, 1.290927710, 0.805102002],
        [0.645835061, 1.452200439, 0.858490216],
        [0.508869537, 1.536239134, 0.890883666],
        [0.412577808, 1.582612762, 0.911745085],
        [0.344114873, 1.610404042, 0.926086795],
        [0.293923498, 1.628354048, 0.936489181],
        [0.255930312, 1.640678700, 0.944357105],
        [0.226334911, 1.649564614, 0.950507042],
        [0.202708200, 1.656225672, 0.955442068],
        [0.183450062, 1.661379149, 0.959487644],
    ]
)


def get_perspective_derivatives(func):
    if func == "log":

        def g(x):
            return -np.log(x)

        def dg(x):
            return -np.reciprocal(x)

        def d2g(x):
            return np.reciprocal(x * x)

        def d3g(x):
            return -np.reciprocal(x * x * x) * 2.0

        def xg(x):
            return -x * np.log(x)

        def dxg(x):
            return -np.log(x) - 1.0

        def d2xg(x):
            return -np.reciprocal(x)

        def d3xg(x):
            return np.reciprocal(x * x)

        def h(x):
            return x * np.log(x)

        def dh(x):
            return np.log(x) + 1.0

        def d2h(x):
            return np.reciprocal(x)

        def d3h(x):
            return -np.reciprocal(x * x)

        def xh(x):
            return x * x * np.log(x)

        def dxh(x):
            return 2.0 * x * np.log(x) + x

        def d2xh(x):
            return 2.0 * np.log(x) + 3.0

        def d3xh(x):
            return 2 * np.reciprocal(x)

    elif isinstance(func, (int, float)):
        a = func
        if a > 0 and a < 1:
            sgn = -1
        elif (a > 1 and a < 2) or (a >= -1 and a < 0):
            sgn = 1
        b = 1.0 - a

        def g(x):
            return sgn * np.power(x, a)

        def dg(x):
            return sgn * np.power(x, a - 1) * a

        def d2g(x):
            return sgn * np.power(x, a - 2) * (a * (a - 1))

        def d3g(x):
            return sgn * np.power(x, a - 3) * (a * (a - 1) * (a - 2))

        def xg(x):
            return sgn * np.power(x, a + 1)

        def dxg(x):
            return sgn * np.power(x, a) * (a + 1)

        def d2xg(x):
            return sgn * np.power(x, a - 1) * ((a + 1) * a)

        def d3xg(x):
            return sgn * np.power(x, a - 2) * ((a + 1) * a * (a - 1))

        def h(x):
            return sgn * np.power(x, b)

        def dh(x):
            return sgn * np.power(x, b - 1) * b

        def d2h(x):
            return sgn * np.power(x, b - 2) * (b * (b - 1))

        def d3h(x):
            return sgn * np.power(x, b - 3) * (b * (b - 1) * (b - 2))

        def xh(x):
            return sgn * np.power(x, b + 1)

        def dxh(x):
            return sgn * np.power(x, b) * (b + 1)

        def d2xh(x):
            return sgn * np.power(x, b - 1) * ((b + 1) * b)

        def d3xh(x):
            return sgn * np.power(x, b - 2) * ((b + 1) * b * (b - 1))

    return {
        "g": (g, dg, d2g, d3g),
        "h": (h, dh, d2h, d3h),
        "xg": (xg, dxg, d2xg, d3xg),
        "xh": (xh, dxh, d2xh, d3xh),
    }



# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics

# __init__.py
from qics.cones.nonnegorthant import NonNegOrthant  # noqa
from qics.cones.possemidefinite import PosSemidefinite  # noqa
from qics.cones.secondorder import SecondOrder  # noqa

from qics.cones.classentr import ClassEntr  # noqa
from qics.cones.classrelentr import ClassRelEntr  # noqa

from qics.cones.quantentr import QuantEntr  # noqa
from qics.cones.quantrelentr import QuantRelEntr  # noqa
from qics.cones.quantcondentr import QuantCondEntr  # noqa
from qics.cones.quantkeydist import QuantKeyDist  # noqa

from qics.cones.opperspectr import OpPerspecTr  # noqa
from qics.cones.opperspecepi import OpPerspecEpi  # noqa
